# Copyright 2021 Matthias Rüdt
#
# This file is part of chemometrics.
#
# chemometrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# chemometrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with chemometrics.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import scipy.sparse as sparse
import scipy.sparse.linalg as splinalg
import matplotlib.pyplot as plt
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import check_array
from sklearn.utils.validation import (FLOAT_DTYPES)
from scipy.optimize import minimize_scalar


def asym_ls(X, y, asym_factor=0.1):
    r"""
    Perform an asymmetric least squares regression.

    A linear regression is performed where either negative (default) or
    positive residuals are more strongly penalized.
    `asym_factor` defines the asymmetry of the regression.

    Parameters
    ----------
    X : (n, m) ndarray
        coefficient matrix
    y : {(n,) ndarray, (n, o) ndarray}
        dependent variables

    Returns
    -------
    beta : (m, o) ndarray
        regression coefficients

    Notes
    -----
    Following equation is solved:

    .. math:: \hat{\beta} = \argmin_\beta (w(y-Xb))^2

    with :math:`w` being a diagonal matrix of weights.
    If :math:`(y-Xb)>0: w_{ii}=asym_factor`
    otherwise :math:`w_{ii} = 1 - asym_factor`

    The problem is solved by iteratively adjusting :math:`w` and using a normal
    least-squares regression [1]_. The alogrithm stops as soon as the weights
    are not adjusted from the previous cycle or the maximum number of cycles
    are exceeded.

    References
    ----------
    .. [1] Hans F.M. Boelens, Reyer J. Dijkstra, Paul H.C. Eilers, Fiona
       Fitzpatrick, Johan A. Westerhuis, New background correction method for
       liquid chromatography with diode array detection, infrared spectroscopic
       detection and Raman spectroscopic detection, J. Chromatogr. A,
       vol. 1057, pp. 21-30, 2004.

    Examples
    --------
    X = np.random.normal(size=[10,3])
    y = np.random.normal(size=[10,1])
    beta = chem.asym_als(X, y)
    """
    max_cycles = 10
    n, m = X.shape
    if y.ndim == 1:
        y = y[:, None]
    o = y.shape[1]
    beta = np.zeros(shape=[m, o])

    # generate solver function
    def solver1d(y1d):
        fun = _asym_ls_y1d(X, y1d, asym_factor=asym_factor,
                           max_cycles=max_cycles)
        return fun
    # iterate over each regression
    beta = np.apply_along_axis(solver1d, 0, y)
    return beta


def _asym_ls_y1d(X, y, asym_factor=0.1, max_cycles=10):
    """
    Asymmetric least-squares on 1d y data.
    """
    if y.ndim == 1:
        y = y[:, None]
    # initialize variables for iterative regression
    w_new = np.ones([X.shape[0], 1])
    w = None
    cycle = 0
    # iterate linear regression until weights converge
    while not np.all(w == w_new) and cycle < max_cycles:
        # update weights
        w = w_new.copy()
        # update variables for weighted regression
        X_scaled = w * X
        y_scaled = w * y
        # solve weighted least squares problem
        beta = np.linalg.lstsq(X_scaled, y_scaled, rcond=-1)[0]
        # calculate new weights
        residuals = y - np.dot(X, beta)
        w_new[residuals > 0] = asym_factor
        w_new[residuals <= 0] = 1 - asym_factor
        # increase counter
        cycle += 1
    return beta[:, 0]


class Emsc(TransformerMixin, BaseEstimator):
    r"""
    Perform extended multiplicative scatter correction (EMSC)

    `Emsc` is a spectral pretreatment which is based on a linear decomposition
    of data into baseline contributions and chemical information. Baseline
    contributions are modelled by polynomial terms up to order `p_order`. The
    chemical information is summarized by the mean spectrum orthogonalized to
    the baseline terms. `Emsc` additionally provides a the functionality for
    orthogonalizing spectra with respect to background information and for
    normalizing the returned spectra.

    Parameters
    ----------
    p_order : int
        Polynoms up to order `p_order` are included for baseline subtraction.
    background : {None (default), (m, o) ndarray}
        Perform additional orthogonalization with respect to background. If
        ``None`` omitted. Otherwise, ``m`` variables x ``o`` background spectra
    normalize : {False (default), True}
        Perform normalization of results
    algorithm : {'als' (default)}
        choice of algorithms for regression. Currently, only asymmetric least
        squares is supported

    Attributes
    ----------
    regressor_ : array of floats
        Matrix of regressor variables for background subtraction.
    coefficients_ : array of floats
        Coefficients of last transform.

    Notes
    -----
    An introduction to EMSC is given in [1]_. Asymmetric least squares
    regression may be looked up at [2]_.

    References
    ----------
    .. [1] Nils Kristian Afseth, Achim Kohler, Extended multiplicative signal
       correction in vibrational spectroscopy, a tutorial, Chemometrics and
       Intelligent Laboratory Systems, vol. 117, pp. 92-99, 2012.
    .. [2] Hans F.M. Boelens, Reyer J. Dijkstra, Paul H.C. Eilers, Fiona
       Fitzpatrick, Johan A. Westerhuis, New background correction method for
       liquid chromatography with diode array detection, infrared spectroscopic
       detection and Raman spectroscopic detection, J. Chromatogr. A,
       vol. 1057, pp. 21-30, 2004.
    """

    def __init__(self, p_order=2, background=None, normalize=False,
                 algorithm='als', asym_factor=0.1):
        self.p_order = p_order
        # convert background to two dimensional array if exists
        if background is not None:
            if background.ndim == 1:
                background = background[:, None]
        self.background = background
        self.normalize = normalize
        self.algorithm = algorithm
        self.asym_factor = asym_factor

    def fit(self, X, y=None):
        r"""
        Calculate regression matrix for later use.

        Parameters
        ----------
        X : (n, m) ndarray
            Data to be pretreated. ``n`` samples x ``m`` variables (typically
            wavelengths)
        y
            Ignored
        """
        X = self._validate_data(X, estimator=self, dtype=FLOAT_DTYPES)

        n_series, n_variables = X.shape
        # generate matrix of baseline polynomials
        baseline = np.zeros([n_variables, self.p_order+1])
        multiplier = np.linspace(-1, 1, num=n_variables)
        for i in range(0, self.p_order+1):
            baseline[:, i] = multiplier ** i
        # matrix for summarizing all factors
        regressor = baseline.copy()

        # if included: prepare background data
        if self.background is not None:
            # orthogonalize background to baseline information
            beta_background = asym_ls(baseline, self.background,
                                      asym_factor=self.asym_factor)
            background_pretreated = self.background - np.dot(baseline,
                                                             beta_background)
            regressor = np.concatenate([regressor, background_pretreated],
                                       axis=1)

        # prepare estimate of chemical information
        X_bar = np.mean(X, axis=0)[:, None]  # mean spectra
        beta_X_bar = asym_ls(regressor, X_bar, asym_factor=self.asym_factor)
        X_bar_pretreated = X_bar - np.dot(regressor, beta_X_bar)
        X_bar_pretreated /= np.sum(X_bar_pretreated) # normalize
        self.regressor_ = np.concatenate((regressor, X_bar_pretreated), axis=1)

        return self

    def transform(self, X, copy=True):
        r"""
        Perform baseline correction by subtracting fit of regresor on `X`

        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            The data used to scale along the features axis.

        """

        X = check_array(X, estimator=self, dtype=FLOAT_DTYPES, copy=copy)
        # perform EMSC on data
        coefficients = asym_ls(self.regressor_, X.T,
                               asym_factor=self.asym_factor)
        X -= np.dot(self.regressor_[:, :-1], coefficients[:-1, :]).T

        if self.normalize:
            X /= coefficients[-1, :][:, None]

        self.coefficients_ = coefficients.T
        return X


class Whittaker(TransformerMixin, BaseEstimator):
    r"""
    Smooth with a Whittaker filter

    `Whittaker` smooths `X` with a Whittaker filter. The filter smooths
    the data with a non-parametric line constraint by its derivative
    smoothness. `penalty` defines the penalty on non-smoothness.
    The whittaker smoother is very efficient and a useful drop-in replacement
    for Savitzky-Golay smoothing.

    Parameters
    ----------
    penalty : float or 'auto' (default)
        Scaling factor of the penalty term for non-smoothness. If 'auto' is
        given, a penalty is estimated based on an algorithmically optimized
        leave-one-out cross validation

    constraint_order : int
        Defines on which order of derivative the constraint acts on.

    deriv : int
        Derivative of the data. Default: 0 - no derivative. Note: deriv should
        always be <= constraint_order.

    Attributes
    ----------
    estimate_penalty : boolean
        `True` if penalty is estimated.

    penalty_ : float
        The applied penalty for smoothing.

    solve1d_ : function
        Solver for smoothing of 1D vector

    Notes
    -----
    `Whittaker` uses sparse matrices for efficiency reasons. `X` may
    however be a full matrix.
    In contrast to the proposed algorithm by Eilers [1]_, no Cholesky
    decomposition is used. The reason is twofold. The Cholesky decomposition
    is not implemented for sparse matrices in Numpy/Scipy. Eilers uses the
    Cholesky decomposition to prevent Matlab from "reordering the sparse
    equation systems for minimal bandwidth". Matlab seems to rely on UMFPACK
    for sparse matrix devision [2]_ which implements column reordering for
    sparsity preservation. As sparse matrix we are working with is square and
    positive-definite, we can rely on the builtin `factorize` method, which
    solves with UMFPACK if installed, otherwise with SuperLU.

    Derivatives are implemented by multiplying the smoothed matrix with a
    (local) difference matrix. This is not explicitly described in [1]_.
    However, the approach is consistent with the underlying idea of the
    Whittaker smoother as the (local) differences are used in the derivation of
    the filter. Note: the derivative should always be smaller equal to the
    constraint order. This is, since the Whittaker filter won't explizitly
    penaltize higher derivative fluctuations than the constraint order.

    References
    ----------
    .. [1] Paul H. Eilers, A perfect smoother, Anal. Chem., vol 75, 14, pp.
       3631-3636, 2003.

    .. [2] UMFPAC, https://en.wikipedia.org/wiki/UMFPACK, accessed
       03.May.2020.
    """

    def __init__(self, penalty='auto', constraint_order=2, deriv=0):
        if penalty == 'auto':
            self.isPenaltyEstimated = True
            self.penalty_ = None
        elif type(penalty) in (int, float):
            self.isPenaltyEstimated = False
            self.penalty_ = penalty
        else:
            raise TypeError('penalty type not correct.')

        self.constraint_order = constraint_order
        self.deriv = deriv

    def fit(self, X, y=None):
        r"""
        Calculate regression matrix for later use.

        Parameters
        ----------
        X : (n, m) ndarray
            Data to be pretreated. ``n`` samples x ``m`` variables (typically
            wavelengths)

        y :
            Ignored
        """
        X = self._validate_data(X, estimator=self, dtype=FLOAT_DTYPES)
        if self.isPenaltyEstimated:
            self._estimate_penalty(X)

        self._fit(X)
        return self

    def _fit(self, X):
        "Fit without argument checks or parameter estimation."
        n_series, n_var = X.shape
        C = _get_whittaker_lhs(n_var, self.penalty_, self.constraint_order)
        self.solve1d_ = splinalg.factorized(C.tocsc())

    def transform(self, X, copy=True):
        r"""
        Do Whittaker smoothing

        Parameters
        ----------
        X : (n, m) ndarray
            Data to be pretreated. ``n`` samples x ``m`` variables (typically
            wavelengths)

        copy : bool (`True` default)
            Whether to genrate a copy of the input file or calculate in place.
        """
        # call to differentation free transform
        X = self._transform(X, copy=copy)
        # differentiate
        if self.deriv > 0:
            X = X @ _sp_diff_matrix(X.shape[1], diff_order=self.deriv).T

        return X

    def _transform(self, X, copy=True):
        """
        Perform the filtering without differentiation
        """
        X = check_array(X, estimator=self, dtype=FLOAT_DTYPES, copy=copy)
        X = np.apply_along_axis(self.solve1d_, 1, X)

        return X

    def score(self, X, y=None):
        r"""
        Calculate cross-validation error of Whittaker filter

        Computes the cross-validation error of a Whittaker filter by a
        leave-one-out cross-validation scheme. The algorithm uses an
        approximation scheme and does not perform the explicit leave-one-out
        cross-validation. Users need should be careful when applying this
        cross-validation scheme to data with autocorrelated noise.
        The algorithm then tends to undersmooth the data.

        Parameters
        ----------
        X : (n, m) ndarray
            Data. ``n`` samples x ``m`` variables (typically
            wavelengths)

        y :
            Ignored

        """
        n_var = X.shape[0]
        z = self._transform(X)
        residuals = z - X
        h_bar = _calc_whittaker_h_bar(n_var, self.penalty_,
                                      self.constraint_order)
        # cross-validation error approximation based on formula proposed by
        # Eiler.
        cv_residuals = residuals / (1 - h_bar)
        error = np.sum(cv_residuals ** 2) / n_var
        return error

    def plot(self, X, logpenalty=[-4, 4]):
        r"""
        Plot CV performance over given range

        Provides an analytical plot of the Whittaker filter score depending on
        the penalty. Each score is the estimated based on a leave-one-out
        approach (see also `score`).
        """
        n_points = 100

        penalties = 10 ** np.linspace(logpenalty[0], logpenalty[1], n_points)
        scores = np.zeros(penalties.shape)
        # store original penalty to restore state
        original_penalty = self.penalty_

        for i in range(n_points):
            self.penalty_ = penalties[i]
            self._fit(X)
            scores[i] = self.score(X)

        plt.plot(penalties, scores)
        plt.xlabel('Penalty')
        plt.ylabel('Score')
        ax = plt.gca()
        ax.semilogx()

        # reset original penalty_
        self.penalty_ = original_penalty
        return ax

    def _obj_fun(self, X, penalty):
        r"Objective funtion for penalty estimation"
        self.penalty_ = 10**penalty
        self._fit(X)
        return self.score(X)

    def _estimate_penalty(self, X):
        r"""
        Estimate optimal penalty based on score

        The penalty of the Whittaker filter is adjusted until the leave-one-out
        error is minimized. The function uses Brent's algorithm and varies
        `penalty_` on a logarithmic scale.
        """

        def obj_fun(log_penalty):
            return self._obj_fun(X, log_penalty)
        bracket = [0, 4]
        res = minimize_scalar(obj_fun, bracket=bracket)

        self.penalty_ = 10**res.x


class AsymWhittaker(TransformerMixin, BaseEstimator):
    r"""
    Background correction `X` with an asymmetric Whittaker filter

    `AsymWhittaker` smooths `X` with an asymmetric Whittaker filter. The filter
    estimates the background by a non-parametric line constraint by its
    derivative smoothness. `penalty` defines the penalty on non-smoothness.

    Parameters
    ----------
    penalty : float
        Scaling factor of the penalty term for non-smoothness.

    constraint_order : int
        Defines on which order of derivative the smoothness constraint acts on.

    asym_factor : float
        Relative weight of negative residuals. Positive residuals obtain
        a weight of `1-asym_factor`. The default value is `0.99`.

    Attributes
    ----------

    background_ : (n, m) ndarray
        Estimated background from last call to `transform`.

    Notes
    -----
    `AsymWhittaker` uses sparse matrices for efficiency reasons. `X` may
    however be a full matrix.
    In contrast to the proposed algorithm by Eilers [1]_, no Cholesky
    decomposition is used. The reason is twofold. The Cholesky decomposition
    is not implemented for sparse matrices in Numpy/Scipy. Eilers uses the
    Cholesky decomposition to prevent Matlab from "reordering the sparse
    equation systems for minimal bandwidth". Matlab seems to rely on UMFPACK
    for sparse matrix devision [2]_ which implements column reordering for
    sparsity preservation. As the sparse matrix we are working with is square
    and positive-definite, we can rely on the builtin `factorize` method, which
    solves with UMFPACK if installed, otherwise with SuperLU.

    References
    ----------
    Application of whittaker smoother to spectroscopic data [1]_.

    .. [1] Paul H. Eilers, A perfect smoother, Anal. Chem., vol 75, 14, pp.
       3631-3636, 2003.
    .. [2] UMFPAC, https://en.wikipedia.org/wiki/UMFPACK, accessed
       03.May.2020.
    """

    def __init__(self, penalty, constraint_order=2, asym_factor=0.99):
        self.penalty = penalty
        self.constraint_order = constraint_order
        self.asym_factor = asym_factor

    def fit(self, X, y=None):
        r"""
        Calculate regression matrix for later use.

        Parameters
        ----------
        X : (n, m) ndarray
            Data to be pretreated. ``n`` samples x ``m`` variables (typically
            wavelengths)

        y :
            Ignored
        """
        X = self._validate_data(X, estimator=self, dtype=FLOAT_DTYPES)

        return self

    def transform(self, X, copy=True):
        r"""
        Do asymmetric Whittaker background subtraction

        Parameters
        ----------
        X : (n, m) ndarray
            Data to be pretreated. ``n`` samples x ``m`` variables (typically
            wavelengths)

        copy : bool (`True` default)
            Whether to genrate a copy of the input file or calculate in place.
        """
        X = check_array(X, estimator=self, dtype=FLOAT_DTYPES, copy=copy)
        self.background_ = np.apply_along_axis(self._solve1d, 1, X)
        return X - self.background_

    def _solve1d(self, x, max_iterations=10):
        # prepare loop for asymmetric least squares
        weights = np.ones(self.n_features_in_)
        weights_old = np.zeros(self.n_features_in_)
        counter = 0
        # loop until optimum or max iterations
        while (np.any(weights != weights_old) & (counter < max_iterations)):
            scaled_x = x * weights
            C = _get_whittaker_lhs(self.n_features_in_, self.penalty,
                                   self.constraint_order, weights=weights)
            solve1d = splinalg.factorized(C.tocsc())
            x_bg = solve1d(scaled_x)
            # prepare for next loop: weights and counter
            residuals = x - x_bg
            weights_old = weights.copy()
            weights[residuals < 0] = self.asym_factor
            weights[residuals >= 0] = 1 - self.asym_factor
            counter += 1
        return x_bg


def _get_whittaker_lhs(n_var, penalty, constraint_order, weights=None):
    r"""
    Return the left matrix for Whittaker filter

    Warning: if weights are used, also right hand side (i.e. unsmoothed data)
    needs to be multiplied by weights.
    """
    D = _sp_diff_matrix(n_var, constraint_order)
    if weights is None:
        lhs = sparse.eye(n_var, format='csc') + penalty * D.transpose().dot(D)
    else:
        lhs = sparse.diags(weights, format='csc') + penalty *\
            D.transpose().dot(D)
    return lhs


def _calc_whittaker_h_bar(n_var, penalty, constraint_order,
                          size_estimator=100):
    r"""
    Calculate estimate of the mean diagonal of the Whittaker smoother matrix
    """
    # reduce size of estimator if necessary
    if n_var < size_estimator:
        size_estimator = n_var

    # the penalty needs to be adjusted for an unbiased estimator
    size_ratio = size_estimator / n_var
    adjusted_penalty = size_ratio ** (2 * constraint_order) * penalty
    # calculate H in the approximation size
    C = _get_whittaker_lhs(size_estimator, adjusted_penalty,
                           constraint_order).toarray()
    rhs = np.eye(size_estimator)
    H = np.linalg.lstsq(C, rhs, rcond=-1)[0]
    return np.mean(np.diag(H))


def _sp_diff_matrix(m, diff_order=1):
    r"""
    Generate a sparse difference matrix used for ``Whittaker``
    """
    E = sparse.eye(m, format='csc')
    for i in range(diff_order):
        E = E[1:, :] - E[:-1, :]
    return E



# Copyright 2021 Matthias Rüdt
#
# This file is part of chemometrics.
#
# chemometrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# chemometrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with chemometrics.  If not, see <https://www.gnu.org/licenses/>.
#
# Additional information on the code history, original copyright holder and
# original license is available in ./license.md


"""
McrAR constraints

All classes need a transform method. Note, unlike sklearn, transform can copy
or overwrite input depending on copy attribute.
"""


from abc import ABC, abstractmethod

import numpy as np

from sklearn.utils import check_array

__all__ = [
    "Constraint",
    "Nonneg",
    "CumsumNonneg",
    "ZeroEndPoints",
    "ZeroCumSumEndPoints",
    "Normalizer",
    "CutBelow",
    "CutAbove",
    "CompressBelow",
    "CompressAbove",
    "ReplaceZeros",
    "Unimodal",
]


class Constraint(ABC):
    """ Abstract class for constraints

    Parameters
    ----------
    copy : bool
        Make copy of input data, A; otherwise, overwrite (if mutable)
    """

    def __init__(self, copy=True):
        self.copy = copy

    @abstractmethod
    def transform(self, A):
        """ Transform A input based on constraint """


class Nonneg(Constraint):
    """
    Non-negativity constraint. All negative entries made 0.

    The non-negativity constraint is typically used if a negative result would
    be unphysical or at least highly unlikely. Typical examples: absolute
    concentrations, absorption spectra.

    Parameters
    ----------
    copy : bool
        Make copy of input data, A; otherwise, overwrite (if mutable)
    """

    def __init__(self, copy=False):
        """ A must be non-negative"""
        super().__init__(copy)

    def transform(self, A):
        """ Apply nonnegative constraint"""
        A = check_array(A, dtype="numeric", copy=self.copy)
        A *= A > 0
        return A


class CumsumNonneg(Constraint):
    """
    Cumulative-Summation non-negativity constraint

    The `CumSumNonneg` constraint enforces the cumulative sum over the data
    to always be positive. This is useful for processing first derivative
    data if for the non-derived data negative entries would be unphysical or at
    least highly unlikely. Typical example: first derivative of an absorption
    spectra

    Parameters
    ----------
    copy : bool
        Make copy of input data, A; otherwise, overwrite (if mutable)
    """

    def __init__(self, axis=1, copy=False):
        """ A must be non-negative"""
        super().__init__(copy)
        self.axis = axis

    def transform(self, A):
        """ Apply cumsum nonnegative constraint"""
        A = check_array(A, dtype="numeric", copy=self.copy)

        A *= np.cumsum(A, self.axis) > 0
        return A


class ZeroEndPoints(Constraint):
    """
    Enforce the endpoints to be zero

    The ZeroEndPoints constraint fits a linear baseline to the data and
    subtracts the baseline from the data. Each component is handled
    independently. If a span is given, first `span` points at each end of the
    data are averaged and only thereafter, the baseline is calculated and
    subtracted.
    The ZeroEndPoints constraint may be useful if we want to prevent MCR from
    modeling background contributions.

    Parameters
    ----------
    copy : bool
        Make copy of input data, A; otherwise, overwrite (if mutable)

    axis : int
        Axis to operate on

    span : int
        Number of pixels along the ends to average.
    """

    def __init__(self, axis=-1, span=1, copy=False):
        """ A must be non-negative"""
        super().__init__(copy)
        if [0, 1, -1].count(axis) != 1:
            raise TypeError("Axis must be 0, 1, or -1")

        self.axis = axis
        self.span = span

    def transform(self, A):
        """ Apply zero endpoints constraint"""

        A = check_array(A, dtype="numeric", copy=self.copy)

        # generate view with relevant axis in first dimension
        if self.axis != 0:
            A = A.T

        pix_vec = np.arange(A.shape[0])
        A_start = A[: self.span, :]
        A_end = A[-self.span:, :]
        delta_x = A_end.mean(axis=0) - A_start.mean(axis=0)
        delta_y = pix_vec[-self.span:].mean() - pix_vec[:self.span].mean()
        slope = delta_x / delta_y
        intercept = (
            A_start - np.dot(pix_vec[: self.span, None], slope[None, :])
        ).mean(axis=0)

        A -= np.dot(pix_vec[:, None], slope[None, :]) + intercept[None, :]

        # restore original array format
        if self.axis != 0:
            A = A.T
        return A


class ZeroCumSumEndPoints(Constraint):
    """
    Enforce certain points in cumulative sum to be near-zero.

    Reduces the cumsum baseline of data by piecewise subtraction of
    constant values between given nodes and the end points of the data. This
    approach is useful for first derivative data with localized peaks and
    no expected offset in the pure spectra. Each peak is delimited by a
    node.

    Parameters
    ----------
    copy : bool
        Make copy of input data, A; otherwise, overwrite (if mutable)
    nodes : list of int
        In addition to end-points, other points to ensure are approximately 0
    axis : int
        Axis to operate on
    """

    def __init__(self, nodes=None, axis=-1, copy=False):
        """ A must be non-negative"""
        super().__init__(copy)

        self.nodes = nodes
        if [0, 1, -1].count(axis) != 1:
            raise TypeError("Axis must be 0, 1, or -1")

        self.axis = axis

    def transform(self, A):
        """ Apply cumsum nonnegative constraint"""
        A = check_array(A, dtype="numeric", copy=self.copy)

        # generate view with relevant axis in first dimension
        if self.axis != 0:
            A = A.T

        if self.nodes:
            self.nodes = set(self.nodes)
        else:
            self.nodes = set()

        # generate a set at least including the endpoints of the data
        self.nodes.update({0, A.shape[0]})
        self.nodes.discard(-1)
        self.nodes = list(self.nodes)

        # subtract constant between nodes
        for num in range(len(self.nodes) - 1):
            n0 = self.nodes[num]
            n1 = self.nodes[num + 1]
            A[n0:n1, :] -= A[n0:n1, :].mean(0)[None, :]

        # restore original array format
        if self.axis != 0:
            A = A.T
        return A


class Normalizer(Constraint):
    """
    Normalization constraint.

    Normalizes the data along the selected axis. This means, that the
    transformed vector subsequently sums to one. If the magnitude of the
    factorized scores and loadings is not set by another constraint,
    Normalizer is necessary for a well-defined MCR problem.

    Parameters
    ----------
    axis : int
        Which axis of input matrix A to apply normalization across.
    fix : list
        Keep fixed entries as-is and normalize the remaining entries such
        that total is 1.
    set_zeros_to_feature : int
        Set all samples which sum-to-zero across axis to 1 for a particular
         feature (See Notes)
    copy : bool
        Make copy of input data, A; otherwise, overwrite (if mutable)

    Notes
    -----

    -   For set_zeros_to_feature, assuming the data represents concentration
         with a matrix [n_samples, n_features] and the axis is across the
         features, for every sample that sums to 0 across axis, would be
         replaced with a vector [n_features] of zeros except at
         set_zeros_to_feature, which would equal 1. I.e., this pixel is
         now pure substance of index value set_zeros_to_feature.


    """

    def __init__(self, axis=-1, fix=None, copy=False):
        """Normalize along axis"""
        super().__init__(copy)
        if fix is None:
            self.fix = fix
        elif isinstance(fix, int):
            self.fix = [fix]
        elif isinstance(fix, (list, tuple)):
            self.fix = fix
        elif isinstance(fix, np.ndarray):
            if np.issubdtype(fix.dtype, np.integer):
                self.fix = fix.tolist()
            else:
                raise TypeError("fix ndarray must be of dtype int")
        else:
            raise TypeError(
                "Parameter fix must be of type None, int, list,",
                "tuple, ndarray"
            )

        if not ((axis == 0) | (axis == 1) | (axis == -1)):
            raise ValueError("Axis must be 0,1, or -1")
        self.axis = axis

    def transform(self, A):
        """ Apply normalization constraint """
        A = check_array(A, dtype=float, copy=self.copy)

        # generate view with relevant axis in first dimension
        if self.axis != 0:
            A = A.T

        if not self.fix:  # No fixed axes
            A /= A.sum(axis=0)[None, :]
        else:  # Fixed axes
            axis_list = np.arange(A.shape[0]).tolist()
            not_fix_locs = [
                v for v in axis_list if self.fix.count(v) == 0
            ]
            scaler = np.ones(A.shape)
            div = A[not_fix_locs, :].sum(axis=0)[None, :]
            div[div == 0] = 1
            sum_fixed = A[self.fix, :].sum(axis=0)[None, :]
            scaler[not_fix_locs, :] = (1 - sum_fixed) / div
            A *= scaler

        # restore original array format
        if self.axis != 0:
            A = A.T

        return A


class ReplaceZeros(Constraint):
    """
    Samples that sum-to-zero across axis are replaced with a vector of 0's
    except for a 1 at feature if a single value. In a concentration context,
    e.g., samples with no concentration are replaced with 100% concentration of
    a set feature. If multiple features given, equal amounts of each feature
    (summing to 1) are used.

    Parameters
    ----------
    axis : int
        Which axis of input matrix A to apply normalization acorss.
    feature : int, list, tuple
        Set all samples which sum-to-zero across axis to fval for a particular
         feature (or fractional) for multiple features.
    fval : float
        Value of summation across axis of replacement vector.
    copy : bool
        Make copy of input data, A; otherwise, overwrite (if mutable)

    """

    def __init__(self, axis=-1, feature=None, fval=1, copy=False):
        """Replace sum-to-zero samples with new feature vector along axis"""
        super().__init__(copy)
        self.fval = fval
        if feature is None:
            self.feature = feature
        elif isinstance(feature, int):
            self.feature = [feature]
        elif isinstance(feature, (list, tuple)):
            self.feature = feature
        elif isinstance(feature, np.ndarray):
            if np.issubdtype(feature.dtype, np.integer):
                self.feature = feature.tolist()
            else:
                raise TypeError("ndarrays must be of dtype int")
        else:
            raise TypeError(
                "Parameter feature must be of type None, int,"
                + "list, tuple, ndarray"
            )

        if not ((axis == 0) | (axis == 1) | (axis == -1)):
            raise ValueError("Axis must be 0,1, or -1")
        self.axis = axis

    def transform(self, A):
        """ Apply constraint """
        A = check_array(A, dtype=float, copy=self.copy)

        # generate view with relevant axis in first dimension
        if self.axis != 0:
            A = A.T

        if self.feature:
            replacement = np.zeros(A.shape[self.axis])
            replacement[self.feature] = self.fval
            replacement /= replacement.sum()
            replacement *= self.fval

            A[:, A.sum(axis=0) == 0] = replacement[:, None]

        # restore original array format
        if self.axis != 0:
            A = A.T
        return A


class _CutExclude(Constraint):
    """
    Parent class for methods that cut and can exclude

    Parameters
    ----------

    value : float
        Cutoff value
    axis_sumnz : {None, int}
        If not None, cut below value only applied where sum across specified
        axis does not go to 0, i.e. all values cut.
    exclude : int, list, tuple, ndarray
        Exclude rows/columns along axis defined by `exclude_axis`. If an index
        is listed in `exclude`, all entries along given `exclude_axis` are not
        transformed.
    exclude_axis : int
        Along which axis entries are excluded. See also `exclude`
    copy : bool
        Make copy of input data, A; otherwise, overwrite (if mutable)
    """

    def __init__(
        self,
        value=0,
        axis_sumnz=None,
        exclude=None,
        exclude_axis=-1,
        copy=False
    ):
        """ """
        super().__init__(copy)
        self.value = value
        self.axis = axis_sumnz
        self.exclude = exclude
        self.exclude_axis = exclude_axis

        self._excl_mat = None

    def _make_excl_mat(self, A_shape):
        """generate mask which excludes certain entries"""
        X, Y = np.meshgrid(np.arange(A_shape[1]), np.arange(A_shape[0]))
        if self.exclude is None:
            # no exclusions
            self._excl_mat = np.zeros(X.shape, dtype=bool)
        else:
            # select axis along which to exclude
            if self.exclude_axis == 0:
                # exclude all entries along dimension based on index
                self._excl_mat = np.isin(Y, self.exclude)
            else:
                self._excl_mat = np.isin(X, self.exclude)


class CutBelow(_CutExclude):
    """
    Cut values below (and not-equal to) a certain threshold.

    The values of an array are set to zero, if they are below a certain
    threshold. This implies that the values may afterwards be further from the
    threshold than before. A typical example where this functionality may be
    interesting is, if peaks should be localized and the interaction between
    different components minimized.
    Additionally, this constraint provides functionalities to exclude sections
    of the data from being processed.

    Parameters
    ----------

    value : float
        Cutoff value
    axis_sumnz : {None, int}
        If not None, cut below value only applied where sum across specified
        axis does not go to 0, i.e. all values cut.
    exclude : int, list, tuple, ndarray
        Exclude rows/columns along axis defined by `exclude_axis`. If an index
        is listed in `exclude`, all entries along given `exclude_axis` are not
        transformed.
    exclude_axis : int
        Along which axis entries are excluded. See also `exclude`
    copy : bool
        Make copy of input data, A; otherwise, overwrite (if mutable)
    """

    def __init__(
        self,
        value=0,
        axis_sumnz=None,
        exclude=None,
        exclude_axis=-1,
        copy=False
    ):
        """ Initialize """
        super().__init__(
            value=value,
            axis_sumnz=axis_sumnz,
            exclude=exclude,
            exclude_axis=exclude_axis,
            copy=copy,
        )

    def transform(self, A):
        """ Apply cut-below value constraint"""
        A = check_array(A, dtype=float, copy=self.copy)

        if self._excl_mat is None:
            self._make_excl_mat(A.shape)

        # change behavior depending if axis_sumnz was defined or not
        if self.axis is None:
            A *= (A >= self.value) | self._excl_mat
        else:
            A *= (
                np.alltrue(A < self.value, axis=self.axis, keepdims=True)
                + (A >= self.value)
                + self._excl_mat
            )
        return A


class CutAbove(_CutExclude):
    """
    Cut values above (and not-equal to) a certain threshold

    The values of an array are set to zero, if they are above a certain
    threshold. This implies that the values may afterwards be further from the
    threshold than before. This constraint provides functionalities to exclude
    sections of the data from being processed.

    Parameters
    ----------

    value : float
        Cutoff value
    axis_sumnz : {None, int}
        If not None, cut below value only applied where sum across specified
        axis does not go to 0, i.e. all values cut.
    exclude : int, list, tuple, ndarray
        Exclude rows/columns along axis defined by `exclude_axis`. If an index
        is listed in `exclude`, all entries along given `exclude_axis` are not
        transformed.
    exclude_axis : int
        Along which axis entries are excluded. See also `exclude`
    copy : bool
        Make copy of input data, A; otherwise, overwrite (if mutable)
    """

    def __init__(
        self,
        value=0,
        axis_sumnz=None,
        exclude=None,
        exclude_axis=-1,
        copy=False
    ):
        """ """
        super().__init__(
            value=value,
            axis_sumnz=axis_sumnz,
            exclude=exclude,
            exclude_axis=exclude_axis,
            copy=copy,
        )

    def transform(self, A):
        """ Apply cut-above value constraint"""
        A = check_array(A, dtype=float, copy=self.copy)

        if self._excl_mat is None:
            self._make_excl_mat(A.shape)

        # change behavior depending if axis_sumnz was defined or not
        if self.axis is None:
            A *= (A <= self.value) | self._excl_mat
        else:
            A *= (
                np.alltrue(A > self.value, axis=self.axis, keepdims=True)
                + (A <= self.value)
                + self._excl_mat
            )
        return A


class CompressBelow(Constraint):
    """
    Compress values below (and not-equal to) a certain threshold

    The constraint sets all values which are smaller than the threshold value
    to the threshold value.

    Parameters
    ----------

    value : float
        Cutoff value
    copy : bool
        Make copy of input data, A; otherwise, overwrite (if mutable)
    """

    def __init__(self, value=0, copy=False):
        """  """
        super().__init__(copy)
        self.value = value

    def transform(self, A):
        """ Apply compress-below value constraint"""
        A = check_array(A, dtype=float, copy=self.copy)

        temp = self.value * (A < self.value)
        A *= A >= self.value
        A += temp
        return A


class CompressAbove(Constraint):
    """
    Compress values above (and not-equal to) a certain threshold (set to value)

    The constraint sets all values which are larger than the threshold value
    to the threshold value.

    Parameters
    ----------

    value : float
        Cutoff value
    copy : bool
        Make copy of input data, A; otherwise, overwrite (if mutable)
    """

    def __init__(self, value=0, copy=False):
        """  """
        super().__init__(copy)
        self.value = value

    def transform(self, A):
        """ Apply compress-above value constraint"""
        A = check_array(A, dtype=float, copy=self.copy)

        temp = self.value * (A > self.value)
        A *= A <= self.value
        A += temp
        return A


class Unimodal(Constraint):
    """
    Enforce column-wise unimodality of data.

    The constraint ensures that one value per column represents a maximum
    while values before the maximum increase monotonically, values after the
    maximum decrease monotonically. The constraint is useful for e.g.
    chromatographic data where typically unimodal peaks are observed.

    Parameters
    ----------

    copy : bool
        Make copy of input data, A; otherwise, overwrite (if mutable). default:
        False
    columns : None (default) or list
        Which columns of input matrix A to apply unimodality across.
    """

    def __init__(self, copy=False, columns=None):
        super().__init__()

        self.columns = columns
        self.copy = copy

    def transform(self, A):
        """
        Apply unimodal constraint to A
        """
        A = check_array(A, dtype=float, copy=self.copy)

        columns = self.columns
        if columns is None:
            columns = range(A.shape[1])

        temp = np.apply_along_axis(
                self._vtransform_to_unimodal,
                0,
                A[:, columns])
        A[:, columns] = temp
        return A

    def _vtransform_to_unimodal(self, vector):
        """Transform vector to unimodal"""
        diff = np.diff(vector)
        indmax = np.argmax(vector)
        maxval = vector[indmax]

        # make diffs before indmax >=0, after indmax <=0
        # this implies (non-strict) monotonity
        mask_ascending = diff[:indmax] < 0
        mask_descending = diff[indmax:] > 0
        mask = np.hstack((mask_ascending, mask_descending))
        diff[mask] = 0

        # integrate & redefine offset
        vector[1:] = vector[0] + np.cumsum(diff)
        deltamax = maxval - vector[indmax]
        vector += deltamax

        return vector



# Copyright 2021, 2022 Matthias Rüdt
#
# This file is part of chemometrics.
#
# chemometrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# chemometrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with chemometrics.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np


def generate_spectra(n_wl, n_band, bandwidth):
    r"""
    Generate a dummy spectra with n_band

    A dummy spectra is generated based on a given number of peaks. The
    bandwidth, location and height is selected randomly. The center of each
    peak follows a uniform distribution over the range of the spectra.
    Each peaks bandwidth is drawn from a gamma distribution. The gamma
    distribution has a heavy tail resulting in the generation of (also) broad
    peaks. The peak height is Poisson distributed with a mean height of 5. Each
    peak in the artificial spectrum follows a Gaussian shape.

    Parameters
    ----------
    n_wl : int
        number of wavelengths/signals to generate
    n_band : int
        number of bands to generate
    bandwidth : float
        impacts the bandwidth of each band.

    Returns
    -------
    spectra : ndarray (n_wl, )
        artificial spectra
    """
    wl = np.arange(n_wl)
    spectra = np.zeros(n_wl)
    for i in range(n_band):
        center_wl = np.random.choice(wl)
        bandwidth_i = np.random.gamma(shape=bandwidth, scale=bandwidth)
        intensity = np.random.poisson(lam=5)*np.random.normal(loc=1, scale=0.2)
        current_spectra = intensity * _gaussian_fun(wl, center_wl, bandwidth_i)
        spectra += current_spectra
    return spectra


def generate_background(n_wl, rel_lengthscale=0.5, size=1):
    r"""
    Generate dummy background.

    Generate dummy background by drawing samples from a gaussian process.

    Parameters
    ----------
    n_wl : int
        number of wavelengths/signals to generate
    rel_lengthscale : float
        lengths scale of the gaussian process kernel
    size : int
        number of background spectra to generate

    Returns
    -------
    background : (n_wl, size)
        artificial background spectra
    """
    mean = np.zeros(n_wl)
    # use a gaussian kernel based weighting for the covariance matrix
    x = np.arange(n_wl)[:, None]
    dist = x.T - x
    cov = np.exp(-(dist / (n_wl * rel_lengthscale))**2)

    # draw 'size' samples from gaussian process as data
    background = np.random.multivariate_normal(mean, cov, size=size)
    return background.T


def generate_data(n_wl=100, n_samples=100, n_conc=2, noise=0.1):
    """
    Generate artificial spectroscopic XY data without background

    An artificial spectroscopic dataset is generated, which resembles a Raman,
    FTIR acquisition. The spectra are disturbed by white noise. Concentrations
    are uniformly distributed between 0 and 1.

    Parameters
    ----------
    n_wl : int
        number of wavelengths/signals to generate (default: 100)
    n_samples : int
        number of samples generated (default: 100)
    n_conc : int
        number of background spectra to generate
    noise : float
        white noise level in the artificial spectra

    Returns
    -------
    X : (n_samples, n_wl)
        artificial spectra
    Y : (n_samples, n_conc)
        artificial reference data
    """
    Y = np.random.uniform(size=[n_samples, n_conc])
    spectra = np.zeros(shape=[n_wl, n_conc])

    for i in range(n_conc):
        spectra[:, i] = generate_spectra(n_wl, n_wl//20, 1)

    X = Y @ spectra.T + np.random.normal(scale=noise,
                                         size=[n_samples, n_wl])
    return X, Y


def _gaussian_fun(x, mu, sigma):
    r"""
    Generates Gaussian profile
    """
    return np.exp(-((x - mu) / sigma) ** 2 / 2)


def pseudo_voigt_spectra(x, parameter):
    r"""
    Generates vector based on pseudo-Voigt profiles

    The function takes a wavenumber vector (x) and a set of parameters with the
    shape 4xn. It generates a mechanistic approximation a spectrum based on
    pseudo Voigt peaks.


    Parameters
    ----------
    x : (n_wl, 1)
        evaluated wavelengths
    parameters : (4, n_peaks)
        parameter definition for voigt profiles. parameters[0, :] define the
        peak positions. parameters[1, :] is proportional to the peak area.
        parameters[2, :] define the percentage of Lorentzian to
        Gauss profile.
        parameters[3, :] define the half width at half height of the peak.

    Returns
    -------
    spectra : (n_wl, 1)
        spectra with n_peaks evaluated voigt profiles

    Notes
    -----
    The pseudo-Voigt peaks are implemented similarly as described in [3]_.
    The parameters are chosen such that they do only affect one
    statistical property of the profile: The definition of the peak width is
    selected such that the Full Width at Half Maximum (FWHM) does not change
    when changing the relative contributions of the Gaussian/Lorentzian
    profile. Similarly, the peak height is not affected by the other parameters
    (peaks normalized by height not by area!). This approach to the peak height
    was chosen as the absolute peak height is a more intuitive parameter than
    the peak area.

    """
    if x.ndim == 1:
        x = x[:, None]

    alpha = parameter[1, :, None]  # n_peaks x 1
    beta = parameter[2, :, None]  # n_peaks x 1
    gamma2 = (parameter[3, :, None].T**2)  # 1 x n_peaks
    omega = parameter[0, :, None].T  # 1 x n_peaks

    # shift x coordinates
    delta_x2 = (x-omega)**2  # n_wl x n_peaks

    # calculate matrix of Gaussian peaks (n_wl x n_peaks)
    gaussian_matrix = np.exp(-np.log(2) * delta_x2/gamma2)

    # calculate matrix of Lorentzian peaks (n_wl x n_peaks)
    lorentzian_matrix = gamma2/(gamma2+delta_x2)

    # convert to vector by matrix multiplication w scaling factor
    gaussian_vector = gaussian_matrix @ (alpha * beta)
    lorentzian_vector = lorentzian_matrix @ (alpha * (1-beta))

    return gaussian_vector + lorentzian_vector



# Copyright 2022 Matthias Rüdt
#
# This file is part of chemometrics.
#
# chemometrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# chemometrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with chemometrics.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from scipy import stats
import matplotlib.pyplot as plt


class LVmixin():
    """
    Mixin for latent variable models

    Latent variable models are projection models which transform the X data
    into a latent structure.

    References
    ----------
    Calculations according to

    .. [Eriksson] L. Eriksson, E. Johansson, N. Kettaneh-Wold, J. Trygg, C.
           Wikström, and S. Wold. Multi- and Megavariate Data Analysis,
           Part I Basic Principles and Applications. Second Edition.
    """

    @property
    def _n_components(self):
        """
        Get number of components independent of naming scheme
        """
        try:
            n_comps = self.n_components_
        except AttributeError:
            n_comps = self.n_components
        return n_comps

    def _calculate_x_residual_std_(self, X):
        r"""
        Calculate the standard deviation of the X residuals

        The x residual standard deviation is calculated according to
        [Eriksson]_ not including the correction factor v (since it is not
        exactly defined in [Eriksson]_).
        """
        X_hat = self.inverse_transform(self.transform(X))
        sse_cal = np.sum((X - X_hat)**2)
        norm_factor = (self.n_samples_ - self.n_components
                       - 1) * (X.shape[1] - self.n_components)
        return np.sqrt(sse_cal / norm_factor)

    def dmodx(self, X, normalize=True, absolute=False):
        r"""
        Calculate distance to model hyperplane in X (DModX)

        DModX provides the distance to the model hyperplane spanned by the
        loading vectors. Any information in the predictors that is not captured
        by the PLS model contributes to DModX. If the DModX is normalized,
        DModX is devided by the mean residual variance of X observed during
        model calibration.

        Parameters
        ----------
        X : (n, m) ndarray
            matrix of predictors. n samples x m predictors

        normalize : {True (default); False}
            normalization of DModX by error in X during calibration

        absolute : {True; False (default)}
            return the absolute distance to the model plane (not normalized by
            degrees of freedom)

        Returns
        -------
        dmodx : (n, ) ndarray
            distance of n samples to model hyperplane

        """

        sse = np.sum((X - self.inverse_transform(self.transform(X)))**2,
                     axis=1)
        dmodx = np.sqrt(sse)

        if not absolute:
            dmodx /= np.sqrt(X.shape[1] - self.n_components)
            if normalize:
                dmodx /= self.x_residual_std_
        return dmodx

    def crit_dmodx(self, confidence=0.95):
        r"""
        Critical distance to hyperplane based on an F2 test

        The critical distance to the model hyperplane is estimated based on
        an F2 distribution. Values above crit_dmodx may be considered outliers.
        dmodx is only approximately F2 distributed [Eriksson]_. It is thus
        worthnoting that the estimated critcal distance is biased. It however
        gives a reasonable indication of points worth investigating.

        """
        degf_cali = self.n_features_in_ - self._n_components - 1
        degf_test = self.n_features_in_ - self._n_components
        f_crit = stats.f.ppf(confidence, degf_test,
                             degf_cali)
        return np.sqrt(f_crit)

    def dhypx(self, X):
        r"""
        Normalized distance on hyperplane

        Provides a distance on the hyperplane, normalized by the distance
        observed during calibration. It can be a useful measure to see whether
        new data is comparable to the calibration data. The normalized dhypx
        is slightly biased towards larger values since the estimated
        `x_residual_std_` is slightly underestimated during model calibration
        [Eriksson]_.

        """
        var_cal = np.var(self.x_scores_, axis=0)
        x_scores2_norm = self.transform(X)**2 / var_cal
        return np.sum(x_scores2_norm, axis=1)

    def crit_dhypx(self, confidence=0.95):
        r"""
        Calculate critical dhypx according to Hotelling's T2

        """
        comp = self._n_components
        samples = self.n_samples_
        f_crit = stats.f.ppf(confidence, self._n_components,
                             samples-self._n_components)
        factor = comp * (samples**2 - 1) / (samples * (samples - comp))
        return f_crit * factor

    def distance_plot(self, X, sample_id=None, confidence=0.95):
        r"""
        Plot distances colinear and orthogonal to model predictor hyperplane

        Generates a figure with two subplots. The subplots provide information
        on how `X` behaves compared to the calibration data. Subplots:
        1) Distance in model hyperplane of predictors. Provides insight into
        the magnitude of variation within the hyperplane compared to the
        calibration data. Large values indicate samples which are outside of
        the calibration space but may be described by linearly scaled latent
        variables.
        2) Distance orthogonal to model hyperplane. Provides insight into the
        magnitude of variation orthogonal to the model hyperplane compared to
        the calibration data. Large values indicate samples which show a
        significant trend not observed in the calibration data.

        """
        plt.figure(figsize=(15, 15))

        if not sample_id:
            sample_id = np.arange(X.shape[0])
        # make plots
        # 1) dhypx
        plt.subplot(211)
        plt.plot(sample_id, self.dhypx(X))
        plt.ylabel('X distance on hyperplane')
        plt.axhline(y=self.crit_dhypx(confidence=confidence))

        plt.subplot(212)
        plt.plot(sample_id, self.dmodx(X))
        plt.axhline(y=self.crit_dmodx(confidence=confidence))
        plt.xlabel('Sample ID')
        plt.ylabel('Distance to X-hyperplane')

        return plt.gcf().axes



# Copyright 2021, 2022 Matthias Rüdt
#
# This file is part of chemometrics.
#
# chemometrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# chemometrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with chemometrics.  If not, see <https://www.gnu.org/licenses/>.

import sklearn
from sklearn.cross_decomposition import PLSRegression as _PLSRegression
from sklearn.pipeline import make_pipeline, Pipeline
from sklearn.model_selection import cross_val_score, KFold
from sklearn.base import (
    BaseEstimator,
    TransformerMixin,
    MultiOutputMixin
)
from sklearn.linear_model import LinearRegression
import numpy as np
from scipy.optimize import least_squares
from scipy.optimize._numdiff import approx_derivative
import matplotlib.pyplot as plt
from .utils import pseudo_voigt_spectra
from .base import LVmixin


class PLSRegression(_PLSRegression, LVmixin):
    r"""
    PLS regression with added chemometric functionality


    References
    ----------
    Calculations according to

    .. [Eriksson] L. Eriksson, E. Johansson, N. Kettaneh-Wold, J. Trygg, C.
           Wikström, and S. Wold. Multi- and Megavariate Data Analysis,
           Part I Basic Principles and Applications. Second Edition.


    """

    def __init__(self, n_components=2, *,
                 max_iter=500, tol=1e-06, copy=True):
        super().__init__(n_components=n_components, scale=False,
                         max_iter=max_iter, tol=tol, copy=copy)

    def fit(self, X, Y):
        super().fit(X, Y)
        self.n_samples_ = self.x_scores_.shape[0]
        self.vip_ = self._calculate_vip()
        self.x_residual_std_ = self._calculate_x_residual_std_(X)
        return self

    def _calculate_vip(self):
        r""" Calculate variable importance in projection (VIP)

        The VIP scores provide a concise summary of how important the different
        predictors are for the model. It summarizes and normalizes the
        information of the loading vector(s). All predictors with VIP-scores
        greater than 1 have a greater influence on the model than if all
        predictors were equally informative. Variable/predictor selection can
        easily be performed by dropping all variables with a score smaller than
        a certain threshold. Typically, 1.0 or 1.5 is used as cut-off. Method
        adapted from [2]_.

        References
        ----------

        .. [2] Mehmood et al. Chemometrics and Intelligent Laboratory Systems
               118 (2012) 62–69.
        """
        ss = np.sum(self.y_loadings_ ** 2, axis=0) *\
            np.sum(self.x_scores_ ** 2, axis=0)
        counter = np.sum(ss * self.x_weights_**2, axis=1)
        denominator = np.sum(ss)
        return np.sqrt(self.n_features_in_ * counter / denominator)

    def hat(self, X):
        r"""
        Calculate the hat (projection) matrix

        Calculate the hat matrix in the X/Y score space. The hat matrix
        :math:`H` projects the observed :math:`Y` onto the predicted
        :math:`\hat Y`. For  obtaining the standard hat matrix, the provided X
        matrix should  correspond to the matrix used during the calibration
        (call to `fit`)  [Eriksson]_.

        Parameters
        ----------
        X : (n, m) ndarray
            Matrix of predictors. n samples x m predictors

        Returns
        -------
        hat : (n, n) ndarray
            Hat matrix, symmetric matrix, n x n samples

        """
        S = self.transform(X)
        return S @ np.linalg.inv(S.T @ S) @ S.T

    def leverage(self, X):
        r"""
        Calculate the statistical leverage

        Calculate the leverage (self-influence of Y) in the X/Y score space.
        For obtaining the standard leverage, the provided X matrix should
        correspond to the matrix used during calibration (call to `fit`).

        Parameters
        ----------
        X : (n, m) ndarray
            Matrix of predictors. n samples x m predictors

        Returns
        -------
        leverage : (n, ) ndarray
            leverage for n samples
        """
        return np.diag(self.hat(X))

    def residuals(self, X, Y, scaling='studentize'):
        r"""
        Calculate (normalized) residuals

        Calculate the (normalized) residuals. The scaling scheme may be
        defined between 'none', 'standardize' and 'studentize'. The normalized
        residuals should only be calculated with the current training set.

        Parameters
        ----------
        X : (n, m) ndarray
            Matrix of predictors. n samples x m predictors
        Y : (n, o) ndarray
            Matrix of responses. n samples x o responses
        scaling : {'none', 'standardize', 'studentize' (default)}
            Define scaling of returned residuals

        Returns
        -------
        residuals : (n, o)
            Matrix of unscaled, standardized or studentized residuals

        Notes
        -----
        The response-wise standard deviation :math:`\sigma_j` is calculated
        according to

        .. math:: \sigma_j = \sqrt{\frac{\sum_i=1^n r_{i,j}^2}{n - p}}.

        Residuals are studentized according to

        .. math:: \hat{r}_i = \frac{r_i}{\sigma\sqrt{(1-h_{ii})}},

        with :math:`\hat{r}_i` being the studentized residuals,
        :math:`r_i` the original residuals and :math:`h_{ii}` the
        leverage.
        """
        Y_pred = self.predict(X)
        residuals = Y_pred - Y
        # internal standard deviation
        std = np.sqrt(np.sum(residuals**2, axis=0)
                      / (X.shape[0] - self.n_components))[:, None].T

        if scaling == 'none':
            scaling_factor = 1
        elif scaling == 'standardize':
            scaling_factor = std
        elif scaling == 'studentize':
            scaling_factor = std * np.sqrt(1 - self.leverage(X)[:, None])
        else:
            raise(TypeError(f'unsupported scaling: {scaling}'))

        return residuals / scaling_factor

    def cooks_distance(self, X, Y):
        r"""
        Calculate Cook's distance from the calibration data

        Parameters
        ----------
        X : (n, m) ndarray
            Matrix of predictors. n samples x m predictors
        Y : (n, o) ndarray
            Matrix of responses. n samples x o responses

        Returns
        -------
        distances : (n, o) ndarray
            List of axis for subplots

        Notes
        -----
        Cooks distance is calculated according to

        .. math::

            D_i = \frac{r_i^2}{p\hat\sigma} \frac{h_{ii}}{(1-h_{ii})^2}

        """
        h = self.leverage(X)
        residuals = self.residuals(X, Y, scaling='none')
        mse = (np.sum(residuals**2, axis=0)
               / (X.shape[0] - self.n_components))[:, None].T
        coefficient1 = h / (1-h)**2
        coefficient2 = residuals**2 / (self.n_components * mse)
        return coefficient1[:, None] * coefficient2

    def plot(self, X, Y):
        r"""
        Displays a figure with 4 common analytical plots for PLS models

        Generates a figure with four subplots providing analytical insights
        into the PLS model. Typically, the calibration data is used for
        the method call. following four subplots are generated:
        1) observed -> predicted. Provides insights into the linearity of the
        data and shows how well the model performes over the model range.
        2) predicted -> studentized residuals. Similar to 1). Useful for
        evaluating the error structure (e.g. homoscedasticity) and detecting
        outliers (studentized residuals > 3)
        3) leverage -> studentized residuals. Provides insights into any data
        points/outliers which strongly affect the model. Optimally, the points
        should be scattered in the center left. The plot includes a limit on
        the Cook's distance of 0.5 and 1 as dashed and solid bordeaux
        lines, respectively.
        4) predictors -> VIP. Provides insights into the predictor importance
        for the model.

        Parameters
        ----------
        X : (n, m) ndarray
            Matrix of predictors. n samples x m predictors
        Y : (n, o) ndarray
            Matrix of responses. n samples x o responses

        Returns
        -------
        axes : list(axis, ...)
            List of axis for subplots

        Notes
        -----
        The residuals are studentized according to

        .. math:: \hat{r}_i = \frac{r_i}{\sqrt{MSE (1-h_{ii)}}}

        The Cook's distance limit is calculated according to

        .. math:: \hat{r}_i = \pm \sqrt{D_{crit} p \frac{(1-h_{ii})}{h_{ii}}}

        with :math:`\hat{r}_i` being the studentized residuals,
        :math:`r_i` the original
        residuals, MSE the mean squared error, :math:`h_{ii}` the leverage,
        :math:`D_{crit}` the critical distance, :math:`p` the number of
        latent variables.
        """
        fig = plt.figure(figsize=(15, 15))
        Y_pred = self.predict(X)
        residuals = self.residuals(X, Y)
        leverage = self.leverage(X)

        # make plots
        # 1) observed vs predicted
        plt.subplot(221)
        plt.scatter(Y, Y_pred)
        plt.axline((np.min(Y), np.min(Y)), (np.max(Y), np.max(Y)),
                   color='k', alpha=0.2)
        plt.xlabel('Observed')
        plt.ylabel('Predicted')

        # 2) predicted vs residuals
        plt.subplot(222)
        plt.scatter(Y_pred, residuals)
        plt.axhline(0, color='k', alpha=0.2)
        plt.xlabel('Predicted')
        plt.ylabel('Studentized residuals')

        # 3) leverage mapped to residuals (lmr) plot
        plt.subplot(223)
        for i in range(residuals.shape[1]):
            plt.scatter(leverage, residuals[:, i], alpha=0.5)

        plt.xlabel('Leverage')
        plt.ylabel('Studentized residuals')
        ax = plt.gca()
        xlim = ax.get_xlim()
        ylim = ax.get_ylim()

        # add cook's distance limit
        def cook(distance, linestyle):
            # prepare data from close to zero to bigger than max leverage
            x = np.linspace(1e-8, np.max(leverage)*1.2)
            lim = np.sqrt(distance*self.n_components * (1-x)/x)
            plt.plot(x, lim, color=[0.4, 0, 0], linestyle=linestyle)
            plt.plot(x, -lim, color=[0.4, 0, 0], linestyle=linestyle)

        cook(0.5, 'dotted')
        cook(1, 'solid')
        # reset axis limits
        plt.xlim(xlim)
        plt.ylim(ylim)

        # 4) VIPs
        plt.subplot(224)
        plt.plot(self.vip_)
        plt.axhline(1, color=np.zeros([3]), alpha=0.2)
        plt.axhline(1.5, color=np.zeros([3]), alpha=0.4)
        plt.xlabel('Predictor')
        plt.ylabel('VIP')

        return fig.axes


def fit_pls(X, Y, pipeline: Pipeline = None, cv_object=None, max_lv=10):
    r"""
    Auto-calibrate PLS model and generate analytical plots

    A PLS model is calibrated based on the maximization of the coefficient of
    determination during cross-validation (Q2). The function provides
    multiple plots for assessing the model quality. The first figure addresses
    the model performance during cross validation and the estimation of optimal
    number of latent variables by showing R2/Q2 values (R^2
    as bars, Q^2 as boxplots based on the individual rotations). The
    second figure shows four subplots with analytical information for the
    optimal model. The plotted figures are: a) observed versus predicted 2)
    predicted versus residuals 3) leverage versus residuals 4) Variable
    importance in projection (VIP) scores.

    Parameters
    ----------
    X : (n, m) ndarray
        Matrix of predictors. n samples x m predictors
    Y : (n, o) ndarray
        Matrix of responses. n samples x o responses
    pipeline : {None, sklearn.pipeline.Pipeline}
        A pipeline object providing a workflow of preprocessing and a
        PLSRegression model. The last entry must be a
        chemometrics.PLSRegression instance.
    cv_object : {None, cv_object}
        An object providing guidance for cross-validation. Typically, it will
        be an instance of an sklearn.model_selection.BaseCrossValidator object.
    max_lv : int
        Number of latent variables up to which the cross-validation score will
        be screened.

    Returns
    -------
    pipeline : Pipeline
        The calibrated model pipeline
    summary : dict
        Summary of the model calibration.
    """
    if not pipeline:
        pipeline = make_pipeline(PLSRegression())
    elif not (isinstance(pipeline, sklearn.pipeline.Pipeline)):
        raise TypeError(
            "pipeline argument provided is of type "
            + "{0} and not of type Pipeline.".format(type(pipeline))
        )
    elif not isinstance(pipeline[-1],
                        PLSRegression):
        # check if pipeline ends with PLSRegression
        raise TypeError(
            "Type of last object provided in pipline is "
            + "{0} but should be a PLSRegression".format(type(pipeline[-1]))
        )

    if not cv_object:
        cv_object = KFold(n_splits=5)

    # perform CV of model up to max_lv
    r2 = []
    q2 = []
    for n_lv in range(1, max_lv+1):
        pipeline[-1].n_components = n_lv
        q2.append(cross_val_score(pipeline, X, Y, cv=cv_object))
        r2.append(pipeline.fit(X, Y).score(X, Y))

    q2 = np.stack(q2).T
    r2 = np.array(r2)

    # plot CV results
    fig_cv = plt.figure(figsize=(6, 6))
    plt.bar(np.arange(1, max_lv+1), r2, alpha=0.5)
    plt.boxplot(q2, positions=np.arange(1, max_lv+1))
    plt.xlabel('Latent variables')
    plt.ylabel('R2, Q2')

    # recover best q2 and adjust model accordingly
    pipeline[-1].n_components = np.argmax(np.median(q2, axis=0)) + 1
    pipeline = pipeline.fit(X, Y)

    # plot PLS performance after preprocessing
    if len(pipeline) > 1:
        X_preprocessed = pipeline[:-1].transform(X)
    else:
        X_preprocessed = X
    pipeline[-1].plot(X_preprocessed, Y)
    fig_model = plt.gcf()

    # arrange return arguments
    analysis = {
        'r2': r2,
        'q2': q2,
        'q2_mean': np.mean(q2, axis=0),
        'q2_median': np.median(q2, axis=0),
        'optimal_lv': pipeline[-1].n_components,
        'figure_cv': fig_cv,
        'figure_model': fig_model
    }

    return pipeline, analysis


class IHM(TransformerMixin, MultiOutputMixin,
          BaseEstimator):
    """
    Indirect Hard Modeling (IHM) without linear regression

    IHM models spectra based on a mechanistic model of multiple
    pure component spectra each consisting of flexible peaks. The spectra are
    described by the peak parameters. For new spectra, the mechanistic spectral
    model is adjusted by a parameter optimization. This allows to correct for
    a variety of effects such as instrument specific shifts or sensor
    variability. IHM returns the parameters of the fitted model.

    Parameters
    ----------
    features : ndarray of shape (n_features, 1), default=None
        feature-related x variable
    peak_parameters : list of ndarrays
        List of peak parameter arrays
    bl_order : int (default: 2)
        Order of background polynome
    spectra_generator : function, default=pseudo_voigt_spectra
        Reference to spectra-generating function
    method : {'LG'}
        Algorithm for spectral fit:
            - 'LG' (default): largest gradient method as descirbed in
                [EKriesten]_.
    gradient_truncation : int (default: 20)
        For peak_parameter fitting, only step along the most important gradient
        directions up to the number of directions given by
        `gradient_truncation`.

    Attributes
    ----------
    n_components_ : int
        Number of components in model
    linearized_breakpoints_ : ndarray
        Vector which indicates at what point different sections of the
        linarized parameter vector end. Structure: (backkground parameters,
        component weights, component shifts, spectra parameters)

    Notes
    -----
    The current optimization strategy follows the largest gradient approach
    described in [EKriesten]_ . To reduce the complexity of the optimization
    problem, first global parameters are optimized (background, spectral shift,
    spectral weights). Peak parameters are optimized one by one depending
    on the gradient size up to a certain number of parameters.

    Nomenclature
    ------------
    features:
        vector of feature or spectral dimension (e.g. wavelength, wavenumber)
    peak:
        a feature-associated effect described by a scaled
        probablity function
    component:
        a chemical species described by a linear combination of peaks
    baseline:
        slowely varying effect not associated to a specific component
    spectra:
        a linear combination of multiple components and baseline effects

    References
    ----------
    Implemented according to
    .. [EKriesten] Kriesten et al. Chemometrics and Intelligent Laboratory
        Systems 91 (2008) 181-193.
    """

    def __init__(self, features, peak_parameters, bl_order=2,
                 spectra_generator=pseudo_voigt_spectra,
                 method='LG',
                 gradient_truncation=20):

        self.features = features
        self.n_components_ = len(peak_parameters)
        self.peak_parameters = np.concatenate(peak_parameters, axis=1)
        self.spectra_generator = spectra_generator
        self.bl_order = bl_order
        self.gradient_truncation = gradient_truncation
        self.method = method

        n_peakparameters = self.peak_parameters.size

        # summarize breaks in peak_parameters matrix
        self._component_breaks = np.array(
            [component.shape[1] for component in peak_parameters]
        ).cumsum()

        # summarize information on total parameters
        self.linearized_breakpoints_ = np.array((
                self.bl_order+1,  # baseline
                self.n_components_,  # component weights
                self.n_components_,  # component shifts
                n_peakparameters  # number of peakparameter
            )).cumsum()

        # prepare polynomial baseline matrix for later use
        self._baseline = np.zeros((self.bl_order+1, self.features.shape[0]))
        multiplier = np.linspace(-1, 1, num=self.features.shape[0])
        for i in range(0, self.bl_order+1):
            self._baseline[i, :] = multiplier ** i

    def fit(self, X, y=None):
        pass

    def transform(self, X, y=None):
        """
        Transform spectra in IHM parameter set
        """
        X_transformed = np.apply_along_axis(self._adjust2spectrum, 1, X)
        return X_transformed

    def _adjust2spectrum(self, spectrum):
        self._current_spectrum = spectrum
        # intialize parameter estimates
        self._weights = np.ones(self.n_components_)
        self._bl = np.zeros(self.bl_order+1)
        self._shifts = np.zeros(self.n_components_)
        self._peak_parameters = self.peak_parameters.copy()

        # fit global parameters
        ini_par = np.concatenate([self._bl, self._weights, self._shifts])
        result = least_squares(self._obj_fun_global_par, ini_par)
        breaks = self.linearized_breakpoints_
        self._bl = result.x[:breaks[0]]
        self._weights = result.x[breaks[0]:breaks[1]]
        self._shifts = result.x[breaks[1]:breaks[2]]

        if self.method == 'LG':
            self._largest_gradient()
        else:
            raise(KeyError(f'Unknown method {self.method}'))

        linearized_parameters = np.concatenate([
            self._bl,
            self._weights,
            self._shifts,
            self._peak_parameters.ravel()
        ])
        return linearized_parameters

    def _largest_gradient(self):
        breaks = self.linearized_breakpoints_
        # select parameters for optimization
        jac = approx_derivative(
            lambda x: np.sum(self._obj_fun_all_peak_parameters(x)**2),
            self.peak_parameters.ravel()
        )
        parameter_sequence = np.argsort(jac)[::-1]
        mask = np.zeros(self.peak_parameters.shape)

        if self.gradient_truncation < parameter_sequence.size:
            mask[parameter_sequence[:self.gradient_truncation]] = 1
        else:
            mask[:] = 1

        # adjust peak positions
        sparsity = np.ones(self.features.shape[0])[:, None]*mask[0, :, None].T
        estimate_pp = least_squares(
            self._obj_fun_peak_position,
            self._peak_parameters[0, :],
            jac_sparsity=sparsity
        )
        self._peak_parameters[0, :] = estimate_pp.x

        # fit baseline, weights
        ini_par = np.concatenate([self._bl, self._weights])
        result = least_squares(self._obj_fun_baseline_weights, ini_par)
        self._bl = result.x[:breaks[0]]
        self._weights = result.x[breaks[0]:breaks[1]]

        # optimize all peak parameters
        sparsity = np.ones(
            self.features.shape[0]
        )[:, None]*mask.ravel()[:, None].T
        estimate_pp = least_squares(
            self._obj_fun_all_peak_parameters,
            self.peak_parameters.ravel(),
            jac_sparsity=sparsity
        )
        self._peak_parameters = np.reshape(estimate_pp.x,
                                           self.peak_parameters.shape)

        # fit baseline, weights
        ini_par = np.concatenate([self._bl, self._weights])
        result = least_squares(self._obj_fun_baseline_weights, ini_par)
        self._bl = result.x[:breaks[0]]
        self._weights = result.x[breaks[0]:breaks[1]]

    def _obj_fun_global_par(self, global_param):
        breaks = self.linearized_breakpoints_
        bl = global_param[:breaks[0]]
        weights = global_param[breaks[0]:breaks[1]]
        shifts = global_param[breaks[1]:breaks[2]]
        spectrum = self._compile_spectrum(bl, weights, shifts,
                                          self._peak_parameters)
        return self._current_spectrum - spectrum

    def _obj_fun_peak_position(self, peak_positions):
        # generate adjusted peak parameter set
        peak_parameters = self._peak_parameters.copy()
        peak_parameters[0, :] = peak_positions
        spectrum = self._compile_spectrum(self._bl, self._weights,
                                          self._shifts, peak_parameters)
        return self._current_spectrum - spectrum

    def _obj_fun_baseline_weights(self, bl_weights):
        breaks = self.linearized_breakpoints_
        bl = bl_weights[:breaks[0]]
        weights = bl_weights[breaks[0]:breaks[1]]
        spectrum = self._compile_spectrum(bl, weights, self._shifts,
                                          self._peak_parameters)
        return self._current_spectrum - spectrum

    def _obj_fun_all_peak_parameters(self, peak_parameters):
        peak_parameters = peak_parameters.reshape(self.peak_parameters.shape)
        spectrum = self._compile_spectrum(self._bl, self._weights,
                                          self._shifts, peak_parameters)
        return self._current_spectrum - spectrum

    def _compile_spectrum(self, bl, weights, shifts, peak_parameters):
        """
        Generate a spectrum based on provided parameter set

        Returns
        -------
        spectra : ndarray (n_components x n_features)
        """

        # generate pure component spectra
        pure_spectra = np.zeros([self.n_components_, self.features.shape[0]])
        start_ind = 0
        for i, end_ind in enumerate(self._component_breaks):
            pure_spectra[i, :] = self.spectra_generator(
                self.features - shifts[i],
                peak_parameters[:, start_ind:end_ind]
            ).T
            start_ind = end_ind
        # generate spectra
        spectra = weights @ pure_spectra
        baseline = bl @ self._baseline
        return spectra + baseline


class IHMRegression(IHM):
    """
    Indirect Hard Modeling (IHM) of spectra with OLS prediction

    IHM models spectra based on a mechanistic description of multiple
    pure component spectra each consisting of a set of peaks. The spectra are
    described by the peak parameters. For new spectra, the mechanistic spectral
    model is adjusted by a parameter optimization. This allows to correct for
    a variety of effects such as instrument specific shifts or sensor
    variability. The weights of the optimized spectral parameter set are used
    for concentration predictions. Concentrations should be given as
    molalities. At low concentrations, the solute concentrations may be
    approximated as mass concentration, molarity, molar fraction, weight
    fraction etc (see Notes).

    Note: The first component spectra is always assumed to be the
    solvent. The predicted molalities for the solvent are always constant and
    should, by definition, be equal to the inverse molar weight 1/M_W [mol/kg].

    Parameters
    ----------
    features : ndarray of shape (n_features, 1), default=None
        feature-related x variable
    peak_parameters : list of ndarrays
        List of peak parameter arrays
    bl_order : int (default: 2)
        Order of background polynome
    spectra_generator : function, default=pseudo_voigt_spectra
        Reference to spectra-generating function
    method : {'LG'}
        Algorithm for spectral fit:
            - 'LG' (default): largest gradient method as descirbed in
                [EKriesten]_.
    gradient_truncation : int (default: 20)
        For peak_parameter fitting, only step along the most important gradient
        directions up to the number of directions given by
        `gradient_truncation`.

    Attributes
    ----------
    n_components_ : int
        Number of components in model
    linearized_breakpoints_ : ndarray
        Vector which indicates at what point different sections of the
        linarized parameter vector end. Structure: (background parameters,
        component weights, component shifts, spectra parameters)
    regressor_ : LinearRegression
        Estimator which converts weights to concentrations

    Notes
    -----
    The current optimization strategy follows the largest gradient approach
    described in [EKriesten]_ . To reduce the complexity of the optimization
    problem, first global parameters are optimized (background, spectral shift,
    spectral weights). Peak parameters are optimized one by one depending
    on the gradient size up to a certain number of parameters.

    Concentration: The optical spectroscopic sensor probes an unknown  volume
    $V$. Each chemical species in the probed volume contributes to the Raman
    signal with a weight $w_i$ proportional to the number of moles in the
    probed volume.
    .. math:: w_i \propto N_i

    We may furthermore expand:
    .. math::
        w_i K_i= N_i
        w_i K_i = N_i

    The solvent is denoted by a subscript $S$. Following the same argument as
    above, we may define the solvent mass in the probed volume
    .. math:: m_S = w_S K'_S = w_S \frac{K_S}{M_{W, S}}

    with :math: `M_{W, S}`, the molar mass of the solvent.

    Molality is given by:
    .. math::
        b_i = \frac{N_i}{m_S}=\frac{w_i K_i}{w_S K'_S}=k_i \frac{w_i}{w_S}

    Nomenclature
    ------------
    features:
        feature or spectral dimension (e.g. wavelength, wavenumber)
    peak:
        a feature-associated effect described by a scaled
        probablity function
    component:
        a chemical species described by a linear combination of peaks
    baseline:
        slowely varying effect not associated to a specific component
    spectra:
        a linear combination of multiple components and baseline effects
    """
    def __init__(self, features, peak_parameters, bl_order=2,
                 spectra_generator=pseudo_voigt_spectra,
                 method='LG',
                 gradient_truncation=20):
        self.regressor_ = LinearRegression(fit_intercept=False)

        super().__init__(features, peak_parameters, bl_order,
                         spectra_generator, method, gradient_truncation)

    def fit(self, X, y):
        """
        Calibrate regression model of IHM
        """
        parameters = self.transform(X)

        weight_range = self.linearized_breakpoints_[0:2]
        weights = parameters[:, weight_range[0]:weight_range[1]]
        normalized_weights = weights / weights[:, 0, None]

        self.regressor_ = self.regressor_.fit(normalized_weights, y)
        return self

    def predict(self, X):
        """
        Predict concentrations from given X
        """
        parameters = self.transform(X)
        weight_range = self.linearized_breakpoints_[0:2]
        weights = parameters[:, weight_range[0]:weight_range[1]]
        normalized_weights = weights / weights[:, 0, None]

        y = self.regressor_.predict(normalized_weights)
        return y



# Copyright 2021 Matthias Rüdt
#
# This file is part of chemometrics.
#
# chemometrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# chemometrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with chemometrics.  If not, see <https://www.gnu.org/licenses/>.
#
# Additional information on the code history, original copyright holder and
# original license is available in ./license.md


"""
Testing chemometrics.mcr.regressors

"""

import numpy as np

from numpy.testing import (assert_equal, assert_array_equal,
                           assert_allclose)

from chemometrics.mcr.regressor import OLS, NNLS

import unittest


class TestRegressors(unittest.TestCase):
    """
    """

    def test_generic_parts(self):
        methods = [OLS, NNLS]

        for method in methods:
            self._test_instantiation(method())
            self._test_dimensionality(method())
            self._test_basic_positive_least_squares(method())

    def _test_instantiation(self, regr):
        """ Test correct instantiation of classes """
        self.assertTrue(regr.coef_ is None)
        self.assertTrue(regr.X_ is None)
        self.assertTrue(regr.residual_ is None)
        self.assertTrue(hasattr(regr, 'fit'))

    def _test_dimensionality(self, regr):
        """ Test correct dimension of regressor methods """

        A = np.array([[1, 1, 0], [1, 0, 1], [0, 0, 1]])
        x = np.array([1, 2, 3])
        X = x[:, None]
        B = np.dot(A, X)
        b = np.dot(A, x)

        # b = 1D
        regr.fit(A, b)
        assert_equal(regr.X_.ndim, 1)
        assert_equal(regr.X_.shape[0], regr.X_.size)
        assert_equal(regr.X_.shape[0], A.shape[-1])
        assert_array_equal(x.shape, regr.X_.shape)
        assert_array_equal(x.T.shape, regr.coef_.shape)

        # B = 2D
        regr.fit(A, B)
        assert_equal(regr.X_.ndim, 2)
        assert_array_equal(regr.X_.shape, [A.shape[-1], B.shape[-1]])

        assert_array_equal(X.shape, regr.X_.shape)
        assert_array_equal(X.T.shape, regr.coef_.shape)

    def _test_basic_positive_least_squares(self, regr):
        """

        """
        A = np.array([[1, 1, 0], [1, 0, 1], [0, 0, 1]])
        x = np.array([1, 2, 3])
        X = x[:, None]
        B = np.dot(A, X)
        b = np.dot(A, x)

        # b is 1D
        regr.fit(A, b)
        assert_allclose(x, regr.X_)

        # b is 1D->2D
        regr.fit(A, b[:, None])
        assert_allclose(X, regr.X_)

        # b = 2D
        regr.fit(A, B)
        assert_allclose(X, regr.X_)

    def test_nnls_negative_x(self):
        """ Test nnls """
        A = np.array([[1, 1, 0], [1, 0, 1], [0, 0, 1]])
        x = np.array([1, -2, 3])
        b = np.dot(A, x)

        regr = NNLS()

        # b is 1D
        regr.fit(A, b)
        assert_allclose(True, regr.X_ >= 0)  # Must be non-negative
        assert_allclose(False, regr.X_ == x)  # Must be non-negative
        # assert_allclose(x, regr.X_)

    def test_ols_negative_x(self):
        """ Test nnls """
        A = np.array([[1, 1, 0], [1, 0, 1], [0, 0, 1]])
        x = np.array([1, -2, 3])
        b = np.dot(A, x)

        regr = OLS()

        # b is 1D
        regr.fit(A, b)
        assert_allclose(x, regr.X_)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 Matthias Rüdt
#
# This file is part of chemometrics.
#
# chemometrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# chemometrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with chemometrics.  If not, see <https://www.gnu.org/licenses/>.

"""
Built-in least squares / regression methods.

All models will follow the formalism, AX = B, solve for X.

NOTE: coef_ will be X.T, which is the formalism that scikit-learn follows

"""

from ._regressor import (
    LinearRegression,
    OLS,
    NNLS
)

__all__ = ['LinearRegression', 'OLS', 'NNLS']



# Copyright 2021 Matthias Rüdt
#
# This file is part of chemometrics.
#
# chemometrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# chemometrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with chemometrics.  If not, see <https://www.gnu.org/licenses/>.

import chemometrics as cm  # accounts for relativ path
from chemometrics import preprocessing as pp
import numpy as np
import unittest
import scipy.sparse as sparse
import matplotlib
import matplotlib.pyplot as plt
from numpy.testing import assert_allclose


class TestAsym_ls(unittest.TestCase):
    """
    Test cases for `asym_ls`
    """

    def test_shape(self):
        """
        Test that output shape is as expected
        """
        shape_x = [10, 3]
        shape_y = [10, 2]
        expected_shape = (shape_x[1], shape_y[1])

        X = np.random.normal(size=shape_x)
        y = np.random.normal(size=shape_y)
        output_shape = cm.asym_ls(X, y).shape
        self.assertEqual(expected_shape, output_shape)

    def test_w_symmetric_ls(self):
        """
        Test if symmetric weights results in least squares solution
        """
        shape_x = [10, 3]
        shape_y = [10, 1]
        X = np.random.normal(size=shape_x)
        y = np.random.normal(size=shape_y)

        beta_als = cm.asym_ls(X, y, asym_factor=0.5)
        beta_ls = np.linalg.lstsq(X, y, rcond=-1)[0]
        self.assertTrue(np.all(np.isclose(beta_als, beta_ls)))

    def test_location(self):
        """
        Test if different asym_factors produce qualitatively correct effect
        """
        shape = [10, 1]
        asym_factors = [0.001, 0.0011, 0.5, 0.9, 0.99]
        x = np.ones(shape)
        y = np.arange(shape[0])[:, None]
        last_beta = 0

        for factor_i in asym_factors:
            current_beta = cm.asym_ls(x, y, asym_factor=factor_i)
            self.assertTrue(current_beta > last_beta)


class TestEmsc(unittest.TestCase):
    r"""
    Test the `Emsc` preprocessing.
    """

    def test_shape(self):
        r"""
        Check the shape of the return matrix
        """
        n_series, n_variables = (10, 50)
        # generate dummy data and background
        scaler = np.linspace(0, 10, num=n_variables)
        D = np.ones([n_series, n_variables]) * scaler[:, None].T
        background = 0.5 * D[0, :]
        background = background[:, None]
        background_list = [None, background]
        # iterate over different inputs
        for bg in background_list:
            emsc = cm.Emsc(p_order=0, background=bg)
            D_pretreated = emsc.fit_transform(D)
            coefficients = emsc.coefficients_
            self.assertTrue(D_pretreated.shape == (n_series, n_variables))
            self.assertTrue(coefficients.shape[0] == n_series)

    def test_background_subtraction(self):
        r"""
        Test whether background subtraction works
        """
        n_series, n_variables = (10, 50)
        # generate dummy data and background
        scaler = np.arange(n_variables)
        D = np.ones([n_series, n_variables]) * scaler[:, None].T
        background = 0.5 * D[0, :]
        background = background[:, None]
        emsc = cm.Emsc(p_order=0, background=background)
        D_pretreated = emsc.fit_transform(D)
        self.assertTrue(np.all(np.isclose(np.zeros([n_series, n_variables]),
                        D_pretreated)))

    def test_normalization(self):
        r"""
        Test if spectra are normalized
        """
        X, y = cm.generate_data(n_wl=99, n_samples=50, n_conc=2, noise=0.1)

        emsc = cm.Emsc(normalize=True)

        X_pretreated = emsc.fit_transform(X)
        background = np.dot(emsc.coefficients_[:, :-1],
                            emsc.regressor_[:, :-1].T)
        X_reconstructed =  background \
                         + emsc.coefficients_[:, -1][:, None]*X_pretreated

        assert_allclose(X, X_reconstructed)


class Testwhittaker(unittest.TestCase):
    r"""
    Test ``whittaker`` smoother.
    """

    def test_shape(self):
        shape = (100, 50)
        penalty = 100
        diff_order = [1, 2]
        X = np.random.normal(size=shape)
        for i in diff_order:
            whittaker = cm.Whittaker(penalty=penalty, constraint_order=i)
            X_smoothed = whittaker.fit_transform(X)
            self.assertEqual(X_smoothed.shape, shape)

    def test_shape_deriv(self):
        r"""
        Test that size of axis 1 of transformed decreases with derivative
        """
        shape = (100, 50)
        penalty = 100
        deriv_order = range(1, 5)
        constraint_order = 5
        X = np.random.normal(size=shape)
        for i in deriv_order:
            whittaker = cm.Whittaker(
                penalty=penalty,
                constraint_order=constraint_order,
                deriv=i
            )
            X_smoothed = whittaker.fit_transform(X)
            expected_shape = (shape[0], shape[1]-i)
            self.assertEqual(X_smoothed.shape, expected_shape)

    def test_null_smoothing(self):
        r"""
        Test that very weak smoothing returns itself
        """
        shape = (2, 50)
        penalty = 0
        X = np.random.normal(size=shape)
        whittaker = cm.Whittaker(penalty=penalty)
        X_smoothed = whittaker.fit_transform(X)
        self.assertTrue(np.all(np.isclose(X, X_smoothed)))

    def test_max_smoothing(self):
        r"""
        Test that very strong smoothing leads to expected results.
        """
        # first order penality, no derivative -> constant
        shape = (2, 50)
        penalty = 1e9
        diff_order = 1
        X = np.random.normal(size=shape) + np.arange(shape[1])[:, None].T
        whittaker = cm.Whittaker(penalty=penalty, constraint_order=diff_order)
        X_smoothed = whittaker.fit_transform(X)
        self.assertTrue(np.allclose(X_smoothed, X.mean(axis=1)[:, None]))

        # first order penality, first derivative -> zeros
        whittaker.deriv = 1
        X_smoothed = whittaker.transform(X)
        self.assertTrue(np.allclose(X_smoothed, 0, atol=1.e-5))

        # second order penality, first order derivative -> constant
        whittaker.constraint_order = 2
        X_smoothed = whittaker.fit_transform(X)
        self.assertTrue(np.allclose(X_smoothed,
                                    X_smoothed.mean(axis=1)[:, None]))

    def test_sp_diff_matrix(self):
        r"""
        Test the ``whittaker`` helper function ``_sp_diff_matrix``
        """
        mat_size = 10
        difforder_list = [1, 2, 3]
        # test for different diff orders that sparse matrix is returned and
        # the results are the same as from numpy diff.
        for difforder in difforder_list:
            diff_mat = pp._sp_diff_matrix(mat_size, diff_order=difforder)
            self.assertTrue(sparse.isspmatrix_csc(diff_mat))
            full = diff_mat.toarray()
            comp = np.diff(np.eye(mat_size), n=difforder, axis=0)
            is_close = np.isclose(full, comp)
            self.assertTrue(np.all(is_close))

    def test_calc_whittaker_h_bar(self):
        r"""
        Test that ``_calc_whittaker_h_bar`` returns a float.
        """
        n_var = [30, 300]
        penalty = 0.5
        constraint_order = 2
        for var in n_var:
            h_bar = pp._calc_whittaker_h_bar(var, penalty, constraint_order)
            self.assertIsInstance(h_bar, float)

    def test_whittaker_cve(self):
        r"""
        Test whittaker cross validation returns float.
        """
        penalty = 1e5
        n_wl = 200
        n_band = 20
        bandwidth = 1
        X = cm.generate_background(n_wl) + cm.generate_spectra(n_wl, n_band,
                                                               bandwidth)
        X = X.T
        whittaker = cm.Whittaker(penalty=penalty)
        whittaker.fit(X)
        cve = whittaker.score(X)
        self.assertIsInstance(cve, float)

    def test_autofit(self):
        r"""
        Test auto fit function of whittaker.
        """
        n_wl = 200
        n_band = 20
        bandwidth = 1
        n_samples = 50
        S = cm.generate_background(n_wl).T + cm.generate_spectra(n_wl, n_band,
                                                                 bandwidth)
        C = np.random.uniform(size=n_samples)
        X = C[:, None] * S
        X = X + np.random.normal(size=X.shape, scale=0.1)
        whittaker = cm.Whittaker()
        X_smoothed = whittaker.fit_transform(X)
        self.assertIsInstance(whittaker.penalty_, float)
        self.assertEqual(X_smoothed.shape, X.shape)

    def test_plot(self):
        r"""
        Test that plot function runs
        - returns axes
        - a similar minimum is shown as obtained by auto-optimziation
        - the penalty remains the same before and after running the function

        """
        X, _ = cm.generate_data()
        whittaker = cm.Whittaker()

        autoscore = whittaker.fit(X).score(X)
        autopenalty = whittaker.penalty_

        plt.figure()
        ax = whittaker.plot(X)
        self.assertIsInstance(ax, matplotlib.axes.Axes)

        y_plotted = ax.lines[0].get_ydata()
        min_plotted = np.min(y_plotted)
        self.assertTrue(np.abs(autoscore-min_plotted) < .5)
        self.assertFalse(np.allclose(y_plotted, min_plotted))

        self.assertEqual(whittaker.penalty_, autopenalty)


class TestAsymWhittaker(unittest.TestCase):
    r"""
    Test ``AsymWhittaker`` background correction.
    """

    def test_shape(self):
        """
        Test shape of return argument
        """
        shape = (100, 50)
        penalty = 100
        diff_order = [1, 2]
        X = np.random.normal(size=shape)
        for i in diff_order:
            aw = cm.AsymWhittaker(penalty=penalty, constraint_order=i)
            X_smoothed = aw.fit_transform(X)
            self.assertEqual(X_smoothed.shape, shape)

    def test_null_smoothing(self):
        r"""
        Test that very weak background subtraction returns 0.
        """
        shape = (1, 50)
        penalty = 0
        X = np.random.normal(size=shape)
        aw = cm.AsymWhittaker(penalty=penalty)
        X_smoothed = aw.fit_transform(X)
        self.assertTrue(np.all(np.isclose(0, X_smoothed)))

    def test_very_asymmetric_bg(self):
        r"""
        Test that a high asym_factor leads to no negative values.
        """
        shape = (1, 50)
        penalty = 100
        diff_order = 2
        asym_factor = 1 - 1e-12
        X = np.random.normal(size=shape)
        aw = cm.AsymWhittaker(penalty=penalty, constraint_order=diff_order,
                              asym_factor=asym_factor)
        X_smoothed = aw.fit_transform(X)
        close_to_zero = np.isclose(0, X_smoothed, atol=1e-5)
        greater_than_zero = X_smoothed > 0
        self.assertTrue(np.all(close_to_zero | greater_than_zero))



# Copyright 2021 Matthias Rüdt
#
# This file is part of chemometrics.
#
# chemometrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# chemometrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with chemometrics.  If not, see <https://www.gnu.org/licenses/>.
#
# Additional information on the code history, original copyright holder and
# original license is available in ./license.md


"""
The :mod:`chemometrics.mcr` module provides Multivariate Curve Resolution -
Alternating Least Squares (MCR-ALS) functionality
"""

from ._mcr import McrAR
from . import constraint
from . import metric
from . import regressor

__all__ = [
    "McrAR",
    "constraint",
    "metric",
    "regressor"
]



# Copyright 2022 Matthias Rüdt
#
# This file is part of chemometrics.
#
# chemometrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# chemometrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with chemometrics.  If not, see <https://www.gnu.org/licenses/>.

import chemometrics as cm
import numpy as np
import unittest
from chemometrics.tests.test_base import TestLVmixin
from sklearn.pipeline import make_pipeline, Pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import KFold
import matplotlib


class TestPCA(unittest.TestCase, TestLVmixin):
    """
    Test PCA class
    """
    def setUp(self):
        self.n_wl = 50
        self.n_samples = 500
        self.n_tests = 100
        self.n_conc = 2

        self.Y = np.random.normal(size=[self.n_samples, self.n_conc])
        noise = 0.1
        spectra = np.zeros(shape=[self.n_wl, self.n_conc])

        for i in range(self.n_conc):
            spectra[:, i] = cm.generate_spectra(self.n_wl, self.n_wl//20, 1)

        self.X = self.Y @ spectra.T + np.random.normal(
            scale=noise,
            size=[self.n_samples, self.n_wl]
        )
        self.model = cm.PCA()
        self.model = self.model.fit(self.X)

        self.Y_test = np.random.normal(size=[self.n_tests, self.n_conc])
        self.X_test = self.Y_test @ spectra.T + np.random.normal(
            scale=noise,
            size=[self.n_tests, self.n_wl]
        )

    def tearDown(self):
        """
        Clean plots etc
        """
        matplotlib.pyplot.close()


class TestFit_pca(unittest.TestCase):
    """
    Test fit_pca function.
    """

    def test_raise_TypeError_pipeline(self):
        """
        Test if function raises a TypeError if wrong pipeline is provided
        """
        X, Y = cm.generate_data()
        with self.assertRaises(TypeError):
            cm.fit_pca(X, pipeline="string as an example wrong object")

    def test_pipeline(self):
        """
        Test if pipeline is accepted as input argument
        """
        X, Y = cm.generate_data()
        pipeline = make_pipeline(StandardScaler(), cm.PCA())
        model, analysis = cm.fit_pca(X, pipeline=pipeline)

    def test_accept_cv_object(self):
        """
        Test if function accepts a cv object
        """
        X, Y = cm.generate_data()
        n_splits = 7
        model, analysis = cm.fit_pca(X, cv_object=KFold(n_splits=n_splits))
        self.assertTrue(analysis['q2'].shape[0] == n_splits)

    def test_return_type(self):
        """
        Test if functions returns an object of type Pipeline
        """
        max_lv = 11
        X, Y = cm.generate_data()
        # basic example
        self._assert_return_args(cm.fit_pca(X, max_lv=max_lv))

        # with pipeline provided
        pipe = make_pipeline(cm.PCA())
        self._assert_return_args(cm.fit_pca(X, pipeline=pipe))

        # with cross validation
        cv = KFold()
        self._assert_return_args(cm.fit_pca(X, cv_object=cv))

    def _assert_return_args(self, returned_args):
        """
        Assert that the returned arguments of fit_pca are correct
        """
        # first return argument must be a pipeline
        pipeline = returned_args[0]
        self.assertIsInstance(pipeline, Pipeline)
        self.assertIsInstance(pipeline[-1], cm.PCA)

        # second return argument must capture info on calibration/cv in dict
        calibration_info = returned_args[1]
        self.assertIsInstance(calibration_info, dict)
        keys = ['q2', 'r2', 'figure_cv', 'figure_model']
        for key in keys:
            self.assertIn(key, calibration_info)

    def tearDown(self):
        """
        Clean plots etc
        """
        matplotlib.pyplot.close()


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Matthias Rüdt
#
# This file is part of chemometrics.
#
# chemometrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# chemometrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with chemometrics.  If not, see <https://www.gnu.org/licenses/>.

"""
Chemometrics: A chemometrics package for Python
===============================================
chemometrics is a Python module providing chemometric functionality in
the Python ecosystem for scientific computing and machine learning (numpy,
scipy, matplotlib, scikit-learn).

"""


from .preprocessing import *
from .plotting import *
from .utils import *
from .regression import *
from . import mcr
from .decomposition import PCA, fit_pca



# Copyright 2021, 2022 Matthias Rüdt
#
# This file is part of chemometrics.
#
# chemometrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# chemometrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with chemometrics.  If not, see <https://www.gnu.org/licenses/>.

import chemometrics as cm
import numpy as np
from numpy.testing import assert_allclose, assert_equal
import unittest
from sklearn.pipeline import make_pipeline, Pipeline
from sklearn.cross_decomposition import PLSRegression
from sklearn.model_selection import KFold
from sklearn.preprocessing import StandardScaler
import matplotlib
from chemometrics.tests.test_base import TestLVmixin


class TestPLSRegression(unittest.TestCase, TestLVmixin):
    """
    Test adjustments done to PLSRegression child class
    """

    def setUp(self):
        self.n_wl = 50
        self.n_samples = 500
        self.n_tests = 100
        self.n_conc = 2

        self.Y = np.random.normal(size=[self.n_samples, self.n_conc])
        noise = 0.1
        spectra = np.zeros(shape=[self.n_wl, self.n_conc])

        for i in range(self.n_conc):
            spectra[:, i] = cm.generate_spectra(self.n_wl, self.n_wl // 20, 1)

        self.X = self.Y @ spectra.T + np.random.normal(
            scale=noise,
            size=[self.n_samples, self.n_wl]
        )
        self.model = cm.PLSRegression()
        self.model = self.model.fit(self.X, self.Y)

        self.Y_test = np.random.normal(size=[self.n_tests, self.n_conc])
        self.X_test = self.Y_test @ spectra.T + np.random.normal(
            scale=noise,
            size=[self.n_tests, self.n_wl]
        )

    def test_vip_shape(self):
        """
        Test the shape of the vip variable
        """
        self.assertTrue(self.model.vip_.shape == (self.n_wl,))

    def test_vip_value(self):
        """
        Test that sum of squared VIPs == number of X variables (definition!)
        """
        self.assertTrue(np.isclose(np.sum(self.model.vip_ ** 2), self.n_wl))

    def test_hat_shape(self):
        """
        Test the shape of the vip variable
        """
        hat = self.model.hat(self.X)
        self.assertTrue(hat.shape == (self.n_samples, self.n_samples))

    def test_hat_works(self):
        """
        Test that hat works according to its definition i.e. H * Y = Ŷ
        """
        hat = self.model.hat(self.X)
        # mean centered prediction
        Y_hat = self.model.predict(self.X) - np.mean(self.Y, axis=0)
        self.assertTrue(np.allclose(hat @ self.Y, Y_hat))

    def test_hat_symmetric(self):
        """
        Test that hat matrix is symmetric
        """
        hat = self.model.hat(self.X)
        self.assertTrue(np.allclose(hat, hat.T))

    def test_hat_indempotent(self):
        """
        Test that hat matrix is indempotent (hat = hat squared)
        """
        hat = self.model.hat(self.X)
        self.assertTrue(np.allclose(hat, hat @ hat))

    def test_leverage_shape(self):
        """
        Test that leverage provides correct matrix shape
        """
        leverage = self.model.leverage(self.X)
        self.assertTrue(leverage.shape == (self.n_samples,))

    def test_leverage_definition(self):
        """
        Test that leverage is diag of the hat matrix
        """
        leverage = self.model.leverage(self.X)
        hat = self.model.hat(self.X)
        self.assertTrue(np.allclose(leverage, np.diag(hat)))

    def test_residuals_shape(self):
        """
        Test residual function
        """
        residuals = self.model.residuals(self.X, self.Y)
        self.assertTrue(residuals.shape == (self.n_samples, self.n_conc))

    def test_residuals_raise_TypeError(self):

        with self.assertRaises(TypeError):
            self.model.residuals(self.X, self.Y, scaling='bla')

    def test_residuals_scaling(self):
        """
        Test different scaling options of residuals
        """
        residuals_unscld = self.model.residuals(self.X, self.Y, scaling='none')
        residuals_std = self.model.residuals(self.X, self.Y,
                                             scaling='standardize')
        residuals_stud = self.model.residuals(self.X, self.Y,
                                              scaling='studentize')
        std_estimated = residuals_unscld / residuals_std

        std = np.sqrt(np.sum(residuals_unscld ** 2, axis=0)
                      / (self.X.shape[0] - self.model.n_components))[:, None].T

        stud_scaling = residuals_std / residuals_stud
        true_scaling = np.sqrt(1 - self.model.leverage(self.X))
        self.assertTrue(np.allclose(std_estimated, std))
        self.assertTrue(np.allclose(stud_scaling, true_scaling[:, None]))

    def test_plot(self):
        """
        Test that plot generates 4 subplot in figure
        """
        axes = self.model.plot(self.X, self.Y)
        for ax in axes:
            self.assertIsInstance(ax, matplotlib.axes.Axes)

        lines = axes[2].get_lines()
        p = self.model.n_components

        for line in lines:
            h = line.get_xdata()
            weighting = h / (1 - h)
            stud_res = line.get_ydata()
            D_rev = stud_res**2 / p * weighting
            close_to05 = np.allclose(D_rev, 0.5)
            close_to1 = np.allclose(D_rev, 1)
            self.assertTrue(close_to05 or close_to1)

    def test_cooks_distance_return_shape(self):
        "Test that cooks_distance return args has correct shape"

        cook = self.model.cooks_distance(self.X, self.Y)
        self.assertTrue(cook.shape == (self.n_samples, self.n_conc))

    def test_cooks_distance_values(self):
        """
        Test reverse function of cooks distance yields residuals
        """
        cook = self.model.cooks_distance(self.X, self.Y)
        residuals = self.model.residuals(self.X, self.Y, scaling='none')

        h = self.model.leverage(self.X)
        weighting = (1 - h) ** 2 / h
        mse = (np.sum(residuals ** 2, axis=0)
               / (self.X.shape[0] - self.model.n_components))[:, None].T
        inverted_res = np.sqrt(cook * self.model.n_components * mse
                               * weighting[:, None])

        self.assertTrue(np.allclose(np.abs(residuals), inverted_res))

    def tearDown(self):
        """
        Clean plots etc
        """
        matplotlib.pyplot.close()


class TestFit_pls(unittest.TestCase):
    """
    Test fit_pls function.
    """

    def test_raise_TypeError_pipeline(self):
        """
        Test if function raises a TypeError if wrong pipeline is provided
        """
        X, Y = cm.generate_data()
        with self.assertRaises(TypeError):
            cm.fit_pls(X, Y, pipeline="string as an example wrong object")

    def test_pipeline(self):
        """
        Test if pipeline is accepted as input argument
        """
        X, Y = cm.generate_data()
        pipeline = make_pipeline(StandardScaler(), cm.PLSRegression())
        model, analysis = cm.fit_pls(X, Y, pipeline=pipeline)

    def test_accept_cv_object(self):
        """
        Test if function raises a TypeError of wrong CV object is provided
        """
        X, Y = cm.generate_data()

        n_splits = 7
        model, analysis = cm.fit_pls(X, Y, cv_object=KFold(n_splits=n_splits))

        self.assertTrue(analysis['q2'].shape[0] == n_splits)

    def test_return_type(self):
        """
        Test if functions returns an object of type Pipeline
        """
        max_lv = 11
        X, Y = cm.generate_data()
        # basic example
        self._assert_return_args(cm.fit_pls(X, Y, max_lv=max_lv))

        # with pipeline provided
        pipe = make_pipeline(cm.PLSRegression())
        self._assert_return_args(cm.fit_pls(X, Y, pipeline=pipe))

        # with cross validation
        cv = KFold()
        self._assert_return_args(cm.fit_pls(X, Y, cv_object=cv))

    def _assert_return_args(self, returned_args):
        """
        Assert that the returned arguments of fit_pls are correct
        """
        # first return argument must be a pipeline
        pipeline = returned_args[0]
        self.assertIsInstance(pipeline, Pipeline)
        self.assertIsInstance(pipeline[-1], PLSRegression)

        # second return argument must capture info on calibration/cv in dict
        calibration_info = returned_args[1]
        self.assertIsInstance(calibration_info, dict)
        keys = ['q2', 'r2', 'figure_cv', 'figure_model']
        for key in keys:
            self.assertIn(key, calibration_info)

    def tearDown(self):
        """
        Clean plots etc
        """
        matplotlib.pyplot.close()


class Test_IHM(unittest.TestCase):
    """
    Test IHM class
    """

    def setUp(self):
        """
        Initialize a dummy instance of ihm
        """
        self.n_features = 200
        self.feature_vector = np.arange(self.n_features)
        self.ini_parameters = [
            np.array([[50, 1, 0.5, 10], [150, 0.5, 0.2, 5]]).T,
            np.array([[100, 1, 0.5, 15]]).T
        ]
        self.ihm = cm.regression.IHM(self.feature_vector, self.ini_parameters)

        # generate set of target parameters different from starting point
        self.true_param = self.ihm.peak_parameters.copy()
        self.true_param *= 1.1

        self.bl = np.array([1, 1, 0.5])
        self.weights = np.array([0.7, 0.3])
        self.shift = np.zeros(2)

        # generate spectrum & fit ihm model
        self.test_spectrum = self.ihm._compile_spectrum(
            self.bl, self.weights, self.shift, self.true_param
        )
        self.ihm._adjust2spectrum(self.test_spectrum)

    def test_init(self):
        """
        Assert that initalization of IHM runs as expected
        """
        ihm = self.ihm
        n_components = len(self.ini_parameters)
        self.assertTrue(
            self.ihm.peak_parameters.shape == (4, 3))
        assert_equal(self.ihm.features, self.feature_vector)
        self.assertTrue(ihm.bl_order == 2)
        self.assertTrue(ihm.n_components_ == n_components)

        assert_allclose(ihm._component_breaks, np.array([2, 3]))
        linearized_breakpoints = np.array([3, 2, 2, 12]).cumsum()
        assert_allclose(ihm.linearized_breakpoints_, linearized_breakpoints)

        assert_allclose(ihm._baseline.shape, np.array([3, self.n_features]))

    def test_compile_spectrum_shape(self):
        """
        Test _compile spectrum
        """
        bl = np.array([1, 1, 0.001])
        weights = np.ones([2])
        shift = np.zeros([2])

        spectrum = self.ihm._compile_spectrum(
            bl, weights, shift, self.ihm.peak_parameters
        )

        self.assertTrue(spectrum.shape == (self.n_features,))

    def test_adjust2spectrum_baseline(self):
        """
        Assert that _adjust2spectrum baseline is close in dummy case
        """
        # check baseline parameters
        assert_allclose(self.ihm._bl, self.bl, rtol=1e-1)

    def test_adjust2spectrum_peak_parameters(self):
        # check peak positions
        estimated_param = self.ihm._peak_parameters.copy()
        start = 0
        for i, end in enumerate(self.ihm._component_breaks):
            estimated_param[0, start:end] += self.ihm._shifts[i]
            start = end
        assert_allclose(estimated_param[0, :], self.true_param[0, :],
                        rtol=1e-2)

    def test_adjust2spectrum_spectrum(self):
        # check accuracy of estimated spectrum
        estimated_spectrum = self.ihm._compile_spectrum(
            self.ihm._bl, self.ihm._weights, self.ihm._shifts,
            self.ihm._peak_parameters
        )
        assert_allclose(estimated_spectrum, self.test_spectrum, rtol=1e-2)

    def test_transform_shape(self):
        """
        Assert that a matrix of parameters is returned by transform
        """
        n_spectra = 3
        spectra = np.zeros([n_spectra, self.n_features])
        pparam = self.ihm.peak_parameters.copy()
        rng = np.random.default_rng(0)

        for i in range(n_spectra):
            bl = rng.uniform(size=[3])
            weights = rng.uniform(size=[2])
            shifts = rng.uniform(size=[2])
            p_scaler = rng.uniform(low=0.5, high=1.5)
            spectra[i, :] = self.ihm._compile_spectrum(
                bl, weights, shifts, p_scaler * pparam
            )

        transformed = self.ihm.transform(spectra)
        length = self.ihm.linearized_breakpoints_[-1]
        self.assertTrue(transformed.shape == (n_spectra, length))

    def test_raises_unkown_method(self):
        """
        Assert that error is raised if an unkown method is provided to ihm
        """
        with self.assertRaises(KeyError):
            ihm = cm.regression.IHM(self.feature_vector, self.ini_parameters,
                                    method='unkown method')
            ihm.transform(self.test_spectrum[:, None])


class Test_IHMRegression(unittest.TestCase):
    """
    Test IHMRegression class
    """

    def setUp(self):
        """
        Initialize a dummy instance of ihm
        """
        self.n_features = 200
        self.n_samples = 5
        rng = np.random.default_rng(seed=10)

        # prepare and initialize IHMRegression
        self.feature_vector = np.arange(self.n_features)
        self.ini_parameters = [
            np.array([[50, 1, 0.5, 10], [150, 0.5, 0.2, 5]]).T,
            np.array([[100, 1, 0.5, 15]]).T
        ]
        self.ihm = cm.regression.IHMRegression(self.feature_vector, self.ini_parameters)

        # generate set of target parameters for regression
        self.true_param = self.ihm.peak_parameters.copy()
        self.true_param *= 1.1

        # generate test spectra and concentrations
        self.X = np.zeros([self.n_samples, self.n_features])
        # molar weight of solvent = 1 mol/kg
        self.y = np.ones([self.n_samples, len(self.ini_parameters)])
        # concentration of second component varies randomly
        self.y[:, 1] = rng.uniform(high=5, size=[self.n_samples])
        self.weights = np.array([10, 1])

        for i in range(self.n_samples):
            self.bl = rng.uniform(high=0.2, size=3)
            self.shift = np.zeros(2)
            conc_weights = self.y[i, :] * self.weights

            # generate spectrum
            self.X[i, :] = self.ihm._compile_spectrum(
                self.bl, conc_weights, self.shift, self.true_param
            )

    def test_fit_precision(self):
        """
        Test if fit estimates parameters accurately
        """
        ihm = self.ihm.fit(self.X, self.y)
        analytical_coef = np.array([1, self.weights[0]])
        assert_allclose(np.diag(ihm.regressor_.coef_), analytical_coef, rtol=1e-1, atol=0.2)

    def test_predict(self):
        """
        Test concentration prediction on calibration set
        """
        y_pred = self.ihm.fit(self.X, self.y).predict(self.X)
        assert_allclose(y_pred, self.y, rtol=1e-1, atol=0.2)

if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Matthias Rüdt
#
# This file is part of chemometrics.
#
# chemometrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# chemometrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with chemometrics.  If not, see <https://www.gnu.org/licenses/>.
#
# Additional information on the code history, original copyright holder and
# original license is available in ./license.md


""" MCR Main Class for Computation"""

import sys as _sys

import numpy as _np
import logging as _logging

from chemometrics.mcr.regressor import OLS, NNLS
import chemometrics.mcr.constraint as constraint
from chemometrics.mcr.metric import mse

from sklearn.base import BaseEstimator, TransformerMixin

# create logger for mcr.py and set default level
_logger = _logging.getLogger(__name__)
_logger.setLevel(_logging.INFO)


class McrAR(TransformerMixin, BaseEstimator):
    """
    Multivariate Curve Resolution - Alternating Regression

    Multivariate curve resolution factorizes data such that the factors
    maximize the explained data. Additional constraints are given for the
    scores and loadings, such as non-negativeness, baseline restrictions,
    unimodality, etc. Due to its origin in the analysis of chemical
    information, a different nomenclature is widely used in literature and also
    in this implementation. Scores are interpreted as concentration matrices
    ``C_``, loadings are the transposed spectral matrices ``ST_``.


    Parameters
    ----------
    c_regr : str, class
        Instantiated regression class (or string, see Notes) for calculating
        the C matrix

    st_regr : str, class
        Instantiated regression class (or string, see Notes) for calculating
        the S^T matrix

    fit_kwargs : dict
        kwargs sent to fit and fit_transform methods

    c_fit_kwargs : dict
        kwargs sent to c_regr.fit method

    st_fit_kwargs : dict
        kwargs sent to``st_regr.fit`` method

    c_constraints : list
        List of constraints applied to calculation of C matrix

    st_constraints : list
        List of constraints applied to calculation of S^T matrix

    max_iter : int
        Maximum number of iterations. One iteration calculates both C and S^T

    err_fcn : function
        Function to calculate error/differences after each least squares
        calculation (ie twice per iteration). Outputs to err attribute.

    tol_increase : float
        Factor increase to allow in err attribute. Set to 0 for no increase
        allowed. E.g., setting to 1.0 means the err can double per iteration.

    tol_n_increase : int
        Number of consecutive iterations for which the err attribute can
        increase

    tol_err_change : float
        If err changes less than tol_err_change, per iteration, break.

    tol_n_above_min : int
        Number of half-iterations that can be performed without reaching a
        new error-minimum

    Attributes
    ----------
    err : list
        List of calculated errors (from err_fcn) after each least squares (ie
        twice per iteration)

    C_ : ndarray [n_samples, n_targets]
        Most recently calculated C matrix (that did not cause a tolerance
        failure)

    ST_ : ndarray [n_targets, n_features]
        Most recently calculated S^T matrix (that did not cause a tolerance
        failure)

    components_ : ndarray [n_targets, n_features]
        Synonym for ``ST_``, providing sklearn like compatibility

    C_opt_ : ndarray [n_samples, n_targets]
        [Optimal] C matrix for lowest err attribute

    ST_opt_ : ndarray [n_targets, n_features]
        [Optimal] ST matrix for lowest err attribute

    n_iter : int
        Total number of iterations performed

    n_features : int
        Total number of features, e.g. spectral frequencies.

    n_samples : int
        Total number of samples (e.g., pixels)

    n_targets : int
        Total number of targets (e.g., pure analytes)

    n_iter_opt : int
        Iteration when optimal C and ST calculated

    exit_max_iter_reached : bool
        Exited iterations due to maximum number of iteration reached (max_iter
        parameter)

    exit_tol_increase : bool
        Exited iterations due to maximum fractional increase in error metric
        (via err_fcn)

    exit_tol_n_increase : bool
        Exited iterations due to maximum number of consecutive increases in
        error metric (via err fcn)

    exit_tol_err_change : bool
        Exited iterations due to error metric change that is smaller than
        tol_err_change

    exit_tol_n_above_min : bool
        Exited iterations due to maximum number of half-iterations for which
        the error metric increased above the minimum error

    Notes
    -----

    -   Built-in regressor classes (str can be used): OLS (ordinary least
        squares), NNLS (non-negatively constrained least squares). See
        mcr.regressors.
    -   Built-in regressor methods can be given as a string to ``c_regr``,
        ``st_regr``;
        though instantiating an imported class gives more flexibility.
    -   Setting any tolerance to None turns that check off

    """

    def __init__(
        self,
        c_regr=OLS(),
        st_regr=OLS(),
        fit_kwargs={},
        c_fit_kwargs={},
        st_fit_kwargs={},
        c_constraints=[constraint.Nonneg()],
        st_constraints=[constraint.Nonneg()],
        max_iter=50,
        err_fcn=mse,
        tol_increase=0.0,
        tol_n_increase=10,
        tol_err_change=None,
        tol_n_above_min=10,
    ):
        """
        Multivariate Curve Resolution - Alternating Regression
        """

        self.fit_kwargs = fit_kwargs

        self.max_iter = max_iter

        self.tol_increase = tol_increase
        self.tol_n_increase = tol_n_increase
        self.tol_err_change = tol_err_change
        self.tol_n_above_min = tol_n_above_min

        self.err_fcn = err_fcn
        self.err = None

        self.c_constraints = c_constraints
        self.st_constraints = st_constraints

        self.c_regressor = self._check_regr(c_regr)
        self.st_regressor = self._check_regr(st_regr)
        self.c_fit_kwargs = c_fit_kwargs
        self.st_fit_kwargs = st_fit_kwargs

        self.C_ = None
        self.ST_ = None

        self.C_opt_ = None
        self.ST_opt_ = None
        self.n_iter_opt = None

        self.n_iter = None
        self.n_increase = None
        self.n_above_min = None

        self.exit_max_iter_reached = False
        self.exit_tol_increase = False
        self.exit_tol_n_increase = False
        self.exit_tol_err_change = False
        self.exit_tol_n_above_min = False

        # Saving every C or S^T matrix at each iteration
        # Could create huge memory usage
        self._saveall_st = False
        self._saveall_c = False
        self._saved_st = []
        self._saved_c = []

    def _check_regr(self, mth):
        """
            Check regressor method. If acceptable strings, instantiate and
            return object. If instantiated class, make sure it has a fit
            attribute.
        """
        if isinstance(mth, str):
            if mth.upper() == "OLS":
                return OLS()
            elif mth.upper() == "NNLS":
                return NNLS()
            else:
                raise ValueError("{} is unknown. Use NNLS or OLS.".format(mth))
        elif hasattr(mth, "fit"):
            return mth
        else:
            raise ValueError(
                "Input class " "{} does not have a 'fit' method".format(mth)
            )

    @property
    def D_(self):
        """:noindex:"""
        return _np.dot(self.C_, self.ST_)

    @property
    def D_opt_(self):
        return _np.dot(self.C_opt_, self.ST_opt_)

    @property
    def n_features(self):
        if self.ST_ is not None:
            return self.ST_.shape[-1]
        else:
            return None

    @property
    def n_targets(self):
        if self.C_ is not None:
            return self.C_.shape[1]
        else:
            return None

    @property
    def n_samples(self):
        if self.C_ is not None:
            return self.C_.shape[0]
        else:
            return None

    def _ismin_err(self, val):
        """ Is the current error the minimum """
        if len(self.err) == 0:
            return True
        else:
            return [val > x for x in self.err].count(True) == 0

    def fit(
        self,
        D,
        C=None,
        ST=None,
        st_fix=None,
        c_fix=None,
        c_first=True,
        verbose=False,
        post_iter_fcn=None,
        post_half_fcn=None,
    ):
        """
        Find factors for decomposing D with MCR-AR method

        Parameters
        ----------
        D : ndarray
            D matrix

        C : ndarray
            Initial C matrix estimate. Only provide initial C OR S^T.

        ST : ndarray
            Initial S^T matrix estimate. Only provide initial C OR S^T.

        st_fix : list
            The spectral component numbers to keep fixed.

        c_fix : list
            The concentration component numbers to keep fixed.

        c_first : bool
            Calculate C first when both C and ST are provided. ``c_fix`` and
            ``st_fix`` must also be provided in this circumstance.

        verbose : bool
            Log iteration and per-least squares err results. See Notes.

        post_iter_fcn : function
            Function to perform after each iteration

        post_half_fcn : function
            Function to perform after half-iteration

        Notes
        -----

        -   Parameters to fit will SUPERCEDE anything in ``fit_kwargs``, if
            provided during McrAR instantiation.
        -   Note that providing C (or ST) to ``fit_kwargs`` and providing
            ``ST`` (or ``C``)
            to fit or fit_transform will raise an error.
        -   When in doubt, clear ``fit_kwargs`` via ``self.fit_kwargs = {}``
        -   Does not affect ``verbose`` or ``c_first`` parameters
        -   Uses the native Python logging module rather than print statements;
            thus, to see the messages, one will need to log-to-file or stream
            to stdout. More info is available in the docs.

        """

        if verbose:
            _logger.setLevel(_logging.DEBUG)
        else:
            _logger.setLevel(_logging.INFO)

        if self.fit_kwargs:
            temp = self.fit_kwargs.get("C")
            if (temp is not None) & (C is None):
                C = temp

            temp = self.fit_kwargs.get("ST")
            if (temp is not None) & (ST is None):
                ST = temp

            temp = self.fit_kwargs.get("st_fix")
            if (temp is not None) & (st_fix is None):
                st_fix = temp

            temp = self.fit_kwargs.get("c_fix")
            if (temp is not None) & (c_fix is None):
                c_fix = temp

            temp = self.fit_kwargs.get("post_iter_fcn")
            if (temp is not None) & (post_iter_fcn is None):
                post_iter_fcn = temp

            temp = self.fit_kwargs.get("post_half_fcn")
            if (temp is not None) & (post_iter_fcn is None):
                post_half_fcn = temp

        # Ensure only C or ST provided
        if (C is None) & (ST is None):
            raise TypeError("C or ST estimate must be provided")
        elif (C is not None) & (ST is not None) & ((c_fix is None) | (st_fix is None)):
            err_str1 = "Only C or ST estimate must be provided, "
            raise TypeError(err_str1 + "unless c_fix and st_fix are both provided")
        else:
            self.C_ = C
            self.ST_ = ST

        self.n_increase = 0
        self.n_above_min = 0
        self.err = []

        # Both C and ST provided. special_skip_c comes into play below
        both_condition = (self.ST_ is not None) & (self.C_ is not None)

        for num in range(self.max_iter):
            self.n_iter = num + 1

            # Both st and c provided, but c_first is False
            if both_condition & (num == 0) & (not c_first):
                special_skip_c = True
            else:
                special_skip_c = False

            if (self.ST_ is not None) & (not special_skip_c):
                # Debugging feature -- saves every S^T matrix in a list
                # Can create huge memory usage
                if self._saveall_st:
                    self._saved_st.append(self.ST_)

                # * Target is the feature of the regression
                self.c_regressor.fit(self.ST_.T, D.T, **self.c_fit_kwargs)
                C_temp = self.c_regressor.coef_

                # Apply fixed C's
                if c_fix:
                    C_temp[:, c_fix] = self.C_[:, c_fix]

                # Apply c-constraints
                for constr in self.c_constraints:
                    C_temp = constr.transform(C_temp)

                # Apply fixed C's
                if c_fix:
                    C_temp[:, c_fix] = self.C_[:, c_fix]

                D_calc = _np.dot(C_temp, self.ST_)

                err_temp = self.err_fcn(C_temp, self.ST_, D, D_calc)

                if self._ismin_err(err_temp):
                    self.C_opt_ = 1 * C_temp
                    self.ST_opt_ = 1 * self.ST_
                    self.n_iter_opt = num + 1
                    self.n_above_min = 0
                else:
                    self.n_above_min += 1

                if self.tol_n_above_min is not None:
                    if self.n_above_min > self.tol_n_above_min:
                        err_str1 = "Half-iterated {} times since " "min ".format(
                            self.n_above_min
                        )
                        err_str2 = "error. Exiting."
                        _logger.info(err_str1 + err_str2)
                        self.exit_tol_n_above_min = True
                        break

                # Calculate error fcn and check for tolerance increase
                if len(self.err) == 0:
                    self.err.append(1 * err_temp)
                    self.C_ = 1 * C_temp
                elif self.tol_increase is None:
                    self.err.append(1 * err_temp)
                    self.C_ = 1 * C_temp
                elif err_temp <= self.err[-1] * (1 + self.tol_increase):
                    self.err.append(1 * err_temp)
                    self.C_ = 1 * C_temp
                else:
                    err_str1 = (
                        "Error increased above fractional"
                        "ctol_increase (C iter). Exiting"
                    )
                    _logger.info(err_str1)
                    self.exit_tol_increase = True
                    break

                # Check if err went up
                if len(self.err) > 1:
                    if self.err[-1] > self.err[-2]:  # Error increased
                        self.n_increase += 1
                    else:
                        self.n_increase *= 0

                # Break if too many error-increases in a row
                if self.tol_n_increase is not None:
                    if self.n_increase > self.tol_n_increase:
                        out_str1 = "Maximum error increases reached "
                        _logger.info(
                            out_str1 + "({}) (C iter). "
                            "Exiting.".format(self.tol_n_increase)
                        )
                        self.exit_tol_n_increase = True
                        break

                _logger.debug(
                    "Iter: {} (C)\t{}: "
                    "{:.4e}".format(self.n_iter, self.err_fcn.__name__, err_temp)
                )

                if post_half_fcn is not None:
                    post_half_fcn(self.C_, self.ST_, D, D_calc)

            if self.C_ is not None:

                # Debugging feature -- saves every C matrix in a list
                # Can create huge memory usage
                if self._saveall_c:
                    self._saved_c.append(self.C_)

                # * Target is the feature of the regression
                self.st_regressor.fit(self.C_, D, **self.st_fit_kwargs)
                ST_temp = self.st_regressor.coef_.T

                # Apply fixed ST's
                if st_fix:
                    ST_temp[st_fix] = self.ST_[st_fix]

                # Apply ST-constraints
                for constr in self.st_constraints:
                    ST_temp = constr.transform(ST_temp)

                # Apply fixed ST's
                if st_fix:
                    ST_temp[st_fix] = self.ST_[st_fix]

                D_calc = _np.dot(self.C_, ST_temp)

                err_temp = self.err_fcn(self.C_, ST_temp, D, D_calc)

                # Calculate error fcn and check for tolerance increase
                if self._ismin_err(err_temp):
                    self.ST_opt_ = 1 * ST_temp
                    self.C_opt_ = 1 * self.C_
                    self.n_iter_opt = num + 1
                    self.n_above_min = 0
                else:
                    self.n_above_min += 1

                if self.tol_n_above_min is not None:
                    if self.n_above_min > self.tol_n_above_min:
                        err_str1 = "Half-iterated {} times " "since min ".format(
                            self.n_above_min
                        )
                        err_str2 = "error. Exiting."
                        _logger.info(err_str1 + err_str2)
                        self.exit_tol_n_above_min = True
                        break

                if len(self.err) == 0:
                    self.err.append(1 * err_temp)
                    self.ST_ = 1 * ST_temp
                elif self.tol_increase is None:
                    self.err.append(1 * err_temp)
                    self.ST_ = 1 * ST_temp
                elif err_temp <= self.err[-1] * (1 + self.tol_increase):
                    self.err.append(1 * err_temp)
                    self.ST_ = 1 * ST_temp
                else:
                    err_str1 = (
                        "Error increased above fractional "
                        "tol_increase (ST iter). Exiting"
                    )
                    _logger.info(err_str1)
                    self.exit_tol_increase = True
                    break

                # Check if err went up
                if len(self.err) > 1:
                    if self.err[-1] > self.err[-2]:  # Error increased
                        self.n_increase += 1
                    else:
                        self.n_increase *= 0

                # Break if too many error-increases in a row
                if self.tol_n_increase is not None:
                    if self.n_increase > self.tol_n_increase:
                        out_str = "Maximum error increases reached "
                        _logger.info(
                            out_str + "({}) (ST iter). "
                            "Exiting.".format(self.tol_n_increase)
                        )
                        self.exit_tol_n_increase = True
                        break

                _logger.debug(
                    "Iter: {} (ST)\t{}: "
                    "{:.4e}".format(self.n_iter, self.err_fcn.__name__, err_temp)
                )

                if post_half_fcn is not None:
                    post_half_fcn(self.C_, self.ST_, D, D_calc)

                if post_iter_fcn is not None:
                    post_iter_fcn(self.C_, self.ST_, D, D_calc)

            if self.n_iter >= self.max_iter:
                _logger.info("Max iterations reached ({}).".format(num + 1))
                self.exit_max_iter_reached = True
                break

            self.n_iter = num + 1

            # Check if err changed (absolute value), per iteration, less
            #  than abs(tol_err_change)

            if (self.tol_err_change is not None) & (len(self.err) > 2):
                err_differ = _np.abs(self.err[-1] - self.err[-3])
                if err_differ < _np.abs(self.tol_err_change):
                    _logger.info(
                        "Change in err below tol_err_change "
                        "({:.4e}). Exiting.".format(err_differ)
                    )
                    self.exit_tol_err_change = True
                    break

    def transform(self, D, **kwargs):
        """
        Transform a data matrix into concentration/score-values

        The method uses half an interation of a normal MCR-alternating
        regression to decompose a data matrix into concentrations/scores given
        the fitted spectral components. The method may be used to retrieve
        concentrations for novel measurements.

        Parameters
        ----------
        D : ndarray
            D matrix

        Returns
        --------

        C : ndarray
            matrix of concentrations is returned

        """

        self.c_regressor.fit(self.ST_.T, D.T, **self.c_fit_kwargs)
        C = self.c_regressor.coef_

        # Apply c-constraints
        for constr in self.c_constraints:
            C = constr.transform(C)

        return C

    @property
    def components_(self):
        return self.ST_


if __name__ == "__main__":  # pragma: no cover
    # PyMCR uses the Logging facility to capture messaging
    # Sends logging messages to stdout (prints them)
    stdout_handler = _logging.StreamHandler(stream=_sys.stdout)
    stdout_format = _logging.Formatter("%(message)s")
    stdout_handler.setFormatter(stdout_format)
    _logger.addHandler(stdout_handler)

    M = 21
    N = 21
    P = 101
    n_components = 2

    C_img = _np.zeros((M, N, n_components))
    C_img[..., 0] = _np.dot(_np.ones((M, 1)), _np.linspace(0, 1, N)[None, :])
    C_img[..., 1] = 1 - C_img[..., 0]

    St_known = _np.zeros((n_components, P))
    St_known[0, 40:60] = 1
    St_known[1, 60:80] = 2

    C_known = C_img.reshape((-1, n_components))

    D_known = _np.dot(C_known, St_known)

    mcrar = McrAR()
    mcrar.fit(D_known, ST=St_known, verbose=True)
    # assert_equal(1, mcrar.n_iter_opt)
    assert ((mcrar.D_ - D_known) ** 2).mean() < 1e-10
    assert ((mcrar.D_opt_ - D_known) ** 2).mean() < 1e-10

    mcrar = McrAR()
    mcrar.fit(D_known, C=C_known)
    # assert_equal(1, mcrar.n_iter_opt)
    assert ((mcrar.D_ - D_known) ** 2).mean() < 1e-10
    assert ((mcrar.D_opt_ - D_known) ** 2).mean() < 1e-10



# Copyright 2021 Matthias Rüdt
#
# This file is part of chemometrics.
#
# chemometrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# chemometrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with chemometrics.  If not, see <https://www.gnu.org/licenses/>.
#
# Additional information on the code history, original copyright holder and
# original license is available in ./license.md


"""
Metrics used in :mod:`chemometrics.mcr`

All functions must take C, ST, D_actual, D_calculated
"""


def mse(C, ST, D_actual, D_calculated):
    """ Mean square error """
    return ((D_actual - D_calculated)**2).sum()/D_actual.size



# Copyright 2021 Matthias Rüdt
#
# This file is part of chemometrics.
#
# chemometrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# chemometrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with chemometrics.  If not, see <https://www.gnu.org/licenses/>.
#
# Additional information on the code history, original copyright holder and
# original license is available in ./license.md


"""
Built-in least squares / regression methods.

All models will follow the formalism, AX = B, solve for X.

"""

from abc import ABC, abstractmethod

import numpy as np

from scipy.linalg import lstsq
from scipy.optimize import nnls


class LinearRegression(ABC):
    """ Abstract class for linear regression methods """

    def __init__(self):
        self.X_ = None
        self.residual_ = None

    @property
    def coef_(self):
        if self.X_ is None:
            return None
        else:
            return self.X_.T

    @abstractmethod
    def fit(self, A, B):
        """ AX = B, solve for X """


class OLS(LinearRegression):
    """
    Ordinary least squares regression

    AX = B, solve for X (coefficients.T)

    Attributes
    ----------
    coef_ : ndarray
        Regression coefficients (X.T)

    residual_ : ndarray
        Residual (sum-of-squares)

    rank_ : int
        Effective rank of matrix A

    svs_ : ndarray
        Singular values of matrix A

    Notes
    -----
    This is simply a wrapped version of Ordinary Least Squares
    (scipy.linalg.lstsq).

    ``coef_`` is X.T, which is the formalism of scikit-learn

    """

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.rank_ = None
        self.svs_ = None

    def fit(self, A, B):
        """ Solve for X: AX = B"""
        self.X_, self.residual_, self.rank_, self.svs_ = lstsq(A, B)


class NNLS(LinearRegression):
    """
    Non-negative constrained least squares regression

    AX = B, solve for X (coeffients.T)

    Attributes
    ----------
    coef_ : ndarray
        Regression coefficients

    residual_ : ndarray
        Residual (sum-of-squares)

    Notes
    -----
    This is simply a wrapped version of NNLS
    (scipy.optimize.nnls).

    ``coef_`` is ``X.T``, which is the formalism of scikit-learn
    """

    def __init__(self, *args, **kwargs):
        super().__init__()

    def fit(self, A, B):
        """ Solve for X: AX = B"""

        if B.ndim == 2:
            N = B.shape[-1]
        else:
            N = 0

        self.X_ = np.zeros((A.shape[-1], N))
        self.residual_ = np.zeros((N))

        # nnls is Ax = b; thus, need to iterate along
        # columns of B
        if N == 0:
            self.X_, self.residual_ = nnls(A, B)
        else:
            for num in range(N):
                self.X_[:, num], self.residual_[num] = nnls(A, B[:, num])
